# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# Filesystem
from pathlib import Path

# GEM
from geode_gem.engine.utils import generate_identifier

from geode_gem.ui.data import Icons
from geode_gem.ui.utils import on_entry_clear, magic_from_file
from geode_gem.ui.dialog.icons import IconsDialog
from geode_gem.ui.widgets.window import CommonWindow

# GObject
from gi.repository import GdkPixbuf, Gtk, Pango

# Translation
from gettext import gettext as _


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class ConsolePreferences(CommonWindow):

    def __init__(self, parent, console, consoles, emulators):
        """ Constructor

        Parameters
        ----------
        parent : Gtk.Window
            Parent object (Default: None)
        console : gem.engine.console.Console
            Console object
        emulators : list
            Emulator objects storage
        """

        CommonWindow.__init__(self,
                              parent,
                              _("Console"),
                              Icons.Symbolic.GAMING,
                              parent.use_classic_theme)

        # ------------------------------------
        #   Initialize variables
        # ------------------------------------

        # GEM API instance
        self.api = parent.api
        # Console object
        self.console = console

        # Consoles storage
        self.consoles = consoles
        # Emulators storage
        self.emulators = emulators

        # GEM config
        self.config = parent.config

        self.path = None

        self.error = False
        self.file_error = False
        self.emulator_error = False

        self.help_data = {
            "order": [
                "Description",
                "Extensions",
                _("Extensions examples"),
                "Expressions"
            ],
            "Description": [
                _("A console represent a games library. You can specify a "
                  "default emulator which is used by this console and "
                  "extensions which is readable by this emulator.")
            ],
            "Extensions": [
                _("Most of the time, extensions are common between differents "
                  "emulators and represent the console acronym name (example: "
                  "Nintendo NES -> nes)."),
                _("Extensions are split by spaces and must not have the first "
                  "dot (using \"nes\" than \".nes\").")
            ],
            _("Extensions examples"): {
                "Nintendo NES": "nes",
                "Sega Megadrive": "md smd bin 32x md cue"
            },
            "Expressions": [
                _("It's possible to hide specific files from the games list "
                  "with regular expressions.")
            ]
        }

        # ------------------------------------
        #   Prepare interface
        # ------------------------------------

        # Init widgets
        self.__init_widgets()

        # Init packing
        self.__init_packing()

        # Init signals
        self.__init_signals()

        # Start interface
        self.__start_interface()

    def __init_widgets(self):
        """ Initialize interface widgets
        """

        self.set_size(640, -1)

        self.set_resizable(True)

        # ------------------------------------
        #   Grids
        # ------------------------------------

        self.grid_preferences = Gtk.Grid()

        self.grid_ignores = Gtk.Box()
        self.grid_ignores_buttons = Gtk.Box()

        # Properties
        self.grid_preferences.set_row_spacing(6)
        self.grid_preferences.set_column_spacing(12)
        self.grid_preferences.set_column_homogeneous(False)

        self.grid_ignores.set_spacing(12)
        self.grid_ignores.set_homogeneous(False)

        Gtk.StyleContext.add_class(
            self.grid_ignores_buttons.get_style_context(), "linked")
        self.grid_ignores_buttons.set_spacing(-1)
        self.grid_ignores_buttons.set_orientation(Gtk.Orientation.VERTICAL)

        # ------------------------------------
        #   Console options
        # ------------------------------------

        self.label_name = Gtk.Label()
        self.entry_name = Gtk.Entry()

        self.label_icon = Gtk.Label()
        self.entry_icon = Gtk.Entry()

        self.label_folder = Gtk.Label()
        self.file_folder = Gtk.FileChooserButton()

        self.button_console = Gtk.Button()
        self.image_console = Gtk.Image()

        self.label_favorite = Gtk.Label()
        self.switch_favorite = Gtk.Switch()

        self.label_recursive = Gtk.Label()
        self.switch_recursive = Gtk.Switch()

        # Properties
        self.label_name.set_halign(Gtk.Align.END)
        self.label_name.set_valign(Gtk.Align.CENTER)
        self.label_name.set_justify(Gtk.Justification.RIGHT)
        self.label_name.get_style_context().add_class(Gtk.STYLE_CLASS_DIM_LABEL)
        self.label_name.set_text(_("Name"))

        self.entry_name.set_hexpand(True)
        self.entry_name.set_icon_from_icon_name(
            Gtk.EntryIconPosition.SECONDARY, Icons.Symbolic.CLEAR)

        self.label_icon.set_halign(Gtk.Align.END)
        self.label_icon.set_valign(Gtk.Align.CENTER)
        self.label_icon.set_justify(Gtk.Justification.RIGHT)
        self.label_icon.get_style_context().add_class(Gtk.STYLE_CLASS_DIM_LABEL)
        self.label_icon.set_text(_("Icon"))

        self.entry_icon.set_hexpand(True)
        self.entry_icon.set_icon_from_icon_name(
            Gtk.EntryIconPosition.SECONDARY, Icons.Symbolic.CLEAR)

        self.label_folder.set_halign(Gtk.Align.END)
        self.label_folder.set_valign(Gtk.Align.CENTER)
        self.label_folder.set_justify(Gtk.Justification.RIGHT)
        self.label_folder.get_style_context().add_class(
            Gtk.STYLE_CLASS_DIM_LABEL)
        self.label_folder.set_text(_("Games folder"))

        self.file_folder.set_action(Gtk.FileChooserAction.SELECT_FOLDER)
        self.file_folder.set_hexpand(True)

        self.button_console.set_size_request(64, 64)
        self.image_console.set_size_request(64, 64)

        self.label_favorite.set_margin_top(12)
        self.label_favorite.set_label(_("Favorite"))
        self.label_favorite.set_halign(Gtk.Align.END)
        self.label_favorite.set_valign(Gtk.Align.CENTER)
        self.label_favorite.get_style_context().add_class(
            Gtk.STYLE_CLASS_DIM_LABEL)

        self.switch_favorite.set_margin_top(12)
        self.switch_favorite.set_halign(Gtk.Align.START)

        self.label_recursive.set_margin_top(12)
        self.label_recursive.set_no_show_all(True)
        self.label_recursive.set_label(_("Recursive"))
        self.label_recursive.set_halign(Gtk.Align.END)
        self.label_recursive.set_valign(Gtk.Align.CENTER)
        self.label_recursive.get_style_context().add_class(
            Gtk.STYLE_CLASS_DIM_LABEL)

        self.switch_recursive.set_margin_top(12)
        self.switch_recursive.set_no_show_all(True)
        self.switch_recursive.set_halign(Gtk.Align.START)

        # ------------------------------------
        #   Emulator options
        # ------------------------------------

        self.label_emulator = Gtk.Label()

        self.label_default = Gtk.Label()

        self.model_emulators = Gtk.ListStore(GdkPixbuf.Pixbuf, str, str)
        self.combo_emulators = Gtk.ComboBox()

        cell_emulators_icon = Gtk.CellRendererPixbuf()
        cell_emulators_name = Gtk.CellRendererText()

        self.label_extensions = Gtk.Label()
        self.entry_extensions = Gtk.Entry()

        # Properties
        self.label_emulator.set_margin_top(18)
        self.label_emulator.set_hexpand(True)
        self.label_emulator.set_use_markup(True)
        self.label_emulator.set_halign(Gtk.Align.CENTER)
        self.label_emulator.set_markup("<b>%s</b>" % _("Default emulator"))

        self.label_default.set_halign(Gtk.Align.END)
        self.label_default.set_valign(Gtk.Align.CENTER)
        self.label_default.set_justify(Gtk.Justification.RIGHT)
        self.label_default.get_style_context().add_class(
            Gtk.STYLE_CLASS_DIM_LABEL)
        self.label_default.set_text(_("Emulator"))

        self.model_emulators.set_sort_column_id(1, Gtk.SortType.ASCENDING)

        self.combo_emulators.set_hexpand(True)
        self.combo_emulators.set_model(self.model_emulators)
        self.combo_emulators.set_id_column(2)
        self.combo_emulators.pack_start(cell_emulators_icon, False)
        self.combo_emulators.add_attribute(cell_emulators_icon, "pixbuf", 0)
        self.combo_emulators.pack_start(cell_emulators_name, True)
        self.combo_emulators.add_attribute(cell_emulators_name, "text", 1)

        cell_emulators_icon.set_padding(4, 0)

        self.label_extensions.set_halign(Gtk.Align.END)
        self.label_extensions.set_valign(Gtk.Align.CENTER)
        self.label_extensions.set_justify(Gtk.Justification.RIGHT)
        self.label_extensions.get_style_context().add_class(
            Gtk.STYLE_CLASS_DIM_LABEL)
        self.label_extensions.set_text(_("Extensions"))

        self.entry_extensions.set_hexpand(True)
        self.entry_extensions.set_tooltip_text(
            _("Use space to separate extensions"))
        self.entry_extensions.set_placeholder_text(
            _("Use space to separate extensions"))
        self.entry_extensions.set_icon_from_icon_name(
            Gtk.EntryIconPosition.SECONDARY, Icons.Symbolic.CLEAR)

        # ------------------------------------
        #   Ignores options
        # ------------------------------------

        self.scroll_ignores = Gtk.ScrolledWindow()

        self.viewport_ignores = Gtk.Viewport()

        self.label_ignores = Gtk.Label()

        self.image_ignores_add = Gtk.Image()
        self.button_ignores_add = Gtk.Button()

        self.image_ignores_remove = Gtk.Image()
        self.button_ignores_remove = Gtk.Button()

        self.model_ignores = Gtk.ListStore(str)
        self.treeview_ignores = Gtk.TreeView()
        self.column_ignores = Gtk.TreeViewColumn()
        self.cell_ignores = Gtk.CellRendererText()

        # Properties
        self.scroll_ignores.set_shadow_type(Gtk.ShadowType.OUT)
        self.scroll_ignores.add(self.viewport_ignores)
        self.scroll_ignores.set_size_request(-1, 200)
        self.scroll_ignores.set_no_show_all(True)

        self.label_ignores.set_margin_top(18)
        self.label_ignores.set_hexpand(True)
        self.label_ignores.set_use_markup(True)
        self.label_ignores.set_no_show_all(True)
        self.label_ignores.set_halign(Gtk.Align.CENTER)
        self.label_ignores.set_markup(
            "<b>%s</b>" % _("Regular expressions for ignored files"))

        self.image_ignores_add.set_from_icon_name(
            Icons.Symbolic.ADD, Gtk.IconSize.BUTTON)
        self.image_ignores_remove.set_from_icon_name(
            Icons.Symbolic.REMOVE, Gtk.IconSize.BUTTON)

        self.button_ignores_add.set_no_show_all(True)
        self.button_ignores_remove.set_no_show_all(True)

        self.treeview_ignores.set_vexpand(True)
        self.treeview_ignores.set_model(self.model_ignores)
        self.treeview_ignores.set_headers_visible(False)
        self.treeview_ignores.set_grid_lines(Gtk.TreeViewGridLines.HORIZONTAL)

        self.column_ignores.pack_start(self.cell_ignores, True)
        self.column_ignores.add_attribute(self.cell_ignores, "text", 0)

        self.cell_ignores.set_padding(12, 6)
        self.cell_ignores.set_property("editable", True)
        self.cell_ignores.set_property("placeholder_text",
                                       _("Write your regex here..."))
        self.cell_ignores.set_property("ellipsize", Pango.EllipsizeMode.END)

        # ------------------------------------
        #   Advanced mode
        # ------------------------------------

        self.check_advanced = Gtk.CheckButton()

        # Properties
        self.check_advanced.set_label(_("Advanced mode"))

    def __init_packing(self):
        """ Initialize widgets packing in main window
        """

        # Main widgets
        self.pack_start(self.grid_preferences, True, True)
        self.pack_start(self.check_advanced, False, False)

        # Console grid
        self.grid_preferences.attach(self.label_name, 0, 0, 1, 1)
        self.grid_preferences.attach(self.entry_name, 1, 0, 1, 1)

        self.grid_preferences.attach(self.label_icon, 0, 1, 1, 1)
        self.grid_preferences.attach(self.entry_icon, 1, 1, 1, 1)

        self.grid_preferences.attach(self.label_folder, 0, 2, 1, 1)
        self.grid_preferences.attach(self.file_folder, 1, 2, 2, 1)

        self.grid_preferences.attach(self.button_console, 2, 0, 1, 2)

        self.grid_preferences.attach(self.label_favorite, 0, 3, 1, 1)
        self.grid_preferences.attach(self.switch_favorite, 1, 3, 2, 1)

        self.grid_preferences.attach(self.label_recursive, 0, 4, 1, 1)
        self.grid_preferences.attach(self.switch_recursive, 1, 4, 2, 1)

        self.grid_preferences.attach(self.label_emulator, 0, 5, 3, 1)

        self.grid_preferences.attach(self.label_default, 0, 6, 1, 1)
        self.grid_preferences.attach(self.combo_emulators, 1, 6, 2, 1)

        self.grid_preferences.attach(self.label_extensions, 0, 7, 1, 1)
        self.grid_preferences.attach(self.entry_extensions, 1, 7, 2, 1)

        self.grid_preferences.attach(self.label_ignores, 0, 8, 3, 1)
        self.grid_preferences.attach(self.grid_ignores, 0, 9, 3, 1)

        # Console options
        self.button_console.set_image(self.image_console)

        # Ignores
        self.button_ignores_add.add(self.image_ignores_add)
        self.button_ignores_remove.add(self.image_ignores_remove)

        self.treeview_ignores.append_column(self.column_ignores)

        self.grid_ignores_buttons.pack_start(
            self.button_ignores_add, False, False, 0)
        self.grid_ignores_buttons.pack_start(
            self.button_ignores_remove, False, False, 0)

        self.grid_ignores.pack_start(
            self.grid_ignores_buttons, False, False, 0)
        self.grid_ignores.pack_start(
            self.scroll_ignores, True, True, 0)

        self.viewport_ignores.add(self.treeview_ignores)

    def __init_signals(self):
        """ Initialize widgets signals
        """

        self.entry_name.connect("changed", self.__on_entry_update)
        self.entry_name.connect("icon-press", on_entry_clear)

        self.entry_icon.connect("changed", self.__on_icon_update)
        self.entry_icon.connect("icon-press", on_entry_clear)

        self.file_folder.connect("file-set", self.__on_entry_update)

        self.entry_extensions.connect("changed", self.__on_entry_update)
        self.entry_extensions.connect("icon-press", on_entry_clear)

        self.button_console.connect("clicked", self.__on_select_icon)

        self.combo_emulators.connect("changed", self.__on_entry_update)

        self.cell_ignores.connect("edited", self.__on_edited_cell)

        self.button_ignores_add.connect("clicked", self.__on_append_item)
        self.button_ignores_remove.connect("clicked", self.__on_remove_item)

        self.check_advanced.connect("toggled", self.__on_check_advanced_mode)

    def __start_interface(self):
        """ Load data and start interface
        """

        self.add_button(_("Close"), Gtk.ResponseType.CLOSE)
        self.add_button(_("Accept"), Gtk.ResponseType.APPLY, Gtk.Align.END)

        self.add_help(self.help_data)

        self.set_response_sensitive(Gtk.ResponseType.APPLY, False)

        emulators_rows = dict()

        for emulator in self.emulators.values():

            if emulator.exists:

                icon = self.main_parent.get_pixbuf_from_cache(
                    "emulators",
                    22,
                    emulator.id,
                    emulator.icon,
                    use_cache=False)

                if icon is None:
                    icon = self.main_parent.icons_blank.get(22)

                emulators_rows[emulator.id] = self.model_emulators.append(
                    [icon, emulator.name, emulator.id])

        self.combo_emulators.set_wrap_width(int(len(self.model_emulators) / 6))

        # ------------------------------------
        #   Init data
        # ------------------------------------

        if self.console is not None and len(self.console.id) > 0:
            self.entry_name.set_text(self.console.name)

            # Folder
            folder = self.console.path
            if folder is not None and folder.exists():
                self.file_folder.set_current_folder(str(folder))

            # Favorite status
            self.switch_favorite.set_active(self.console.favorite)

            # Recursive status
            self.switch_recursive.set_active(self.console.recursive)

            # Extensions
            self.entry_extensions.set_text(' '.join(self.console.extensions))

            # Icon
            self.path = self.console.icon

            if self.path is not None:
                self.entry_icon.set_text(str(self.path))

                icon = self.main_parent.get_pixbuf_from_cache(
                    "consoles", 64, self.console.id, self.console.icon)

                if icon is None:
                    icon = self.main_parent.icons_blank.get(64)

                self.image_console.set_from_pixbuf(icon)

            # Ignores
            for ignore in self.console.ignores:
                self.model_ignores.append([ignore])

            # Emulator
            emulator = self.console.emulator
            if emulator is not None and emulator.id in emulators_rows:
                self.combo_emulators.set_active_id(emulator.id)

            self.__on_entry_update()

        # ------------------------------------
        #   Advanded mode
        # ------------------------------------

        self.check_advanced.set_active(
            self.config.getboolean("advanced", "console", fallback=False))

        self.__on_check_advanced_mode()

    def save(self):
        """ Save modification
        """

        self.section = self.entry_name.get_text().strip()

        if len(self.section) == 0:
            return None

        data = {
            "id": generate_identifier(self.section),
            "name": self.section
        }

        value = self.file_folder.get_filename()
        if value is not None and len(value) > 0:
            data["path"] = Path(value).expanduser()

        value = self.entry_icon.get_text().strip()
        if len(value) > 0:
            data["icon"] = Path(value).expanduser()

        value = self.entry_extensions.get_text().strip()
        if len(value) > 0:
            data["extensions"] = value.split()

        data["ignores"] = list()
        for row in self.model_ignores:
            element = self.model_ignores.get_value(row.iter, 0)

            if element is not None and len(element) > 0:
                data["ignores"].append(element)

        data["favorite"] = self.switch_favorite.get_active()
        data["recursive"] = self.switch_recursive.get_active()

        data["emulator"] = self.emulators.get(
            self.combo_emulators.get_active_id())

        # Avanced view status
        status = self.config.getboolean("advanced", "console", fallback=False)

        if not self.check_advanced.get_active() == status:
            self.config.modify(
                "advanced", "console", self.check_advanced.get_active())
            self.config.update()

        return data

    def __on_entry_update(self, widget=None):
        """ Update dialog response sensitive status

        Parameters
        ----------
        widget : Gtk.Widget, optional
            Object which receive signal
        """

        self.error = False

        # ------------------------------------
        #   Console name
        # ------------------------------------

        icon = None
        tooltip = None

        name = self.entry_name.get_text()

        if len(name) > 0:

            # Always check identifier to avoid NES != NeS
            name = generate_identifier(name)

            # Check if current console exists in database
            if name in self.consoles.keys():

                if self.console is None:
                    self.error = True

                # Avoid to use a name which already exists in database
                elif not self.console.id == name:
                    self.error = True

                if self.error:
                    icon = Icons.ERROR
                    tooltip = _("This console already exists, please, "
                                "choose another name")

        else:
            self.error = True

        self.entry_name.set_icon_from_icon_name(
            Gtk.EntryIconPosition.PRIMARY, icon)
        self.entry_name.set_tooltip_text(tooltip)

        # ------------------------------------
        #   Console roms folder
        # ------------------------------------

        path = self.file_folder.get_filename()

        if path is None or not Path(path).expanduser().exists():
            self.error = True

        # ------------------------------------
        #   Console roms extensions
        # ------------------------------------

        extensions = self.entry_extensions.get_text().strip()

        if len(extensions) == 0:
            self.error = True

        # ------------------------------------
        #   Console emulator
        # ------------------------------------

        if self.combo_emulators.get_active_id() is None:
            self.error = True

        # ------------------------------------
        #   Start dialog
        # ------------------------------------

        self.set_response_sensitive(Gtk.ResponseType.APPLY, not self.error)

    def __on_icon_update(self, widget):
        """ Update icon thumbnail when the icon entry is update

        Parameters
        ----------
        widget : Gtk.Widget
            Object which receive signal
        """

        value = widget.get_text()

        self.path = None
        if len(value) > 0:
            self.path = Path(value).expanduser()

        self.set_icon(self.image_console, self.path)

    def __on_select_icon(self, widget):
        """ Select a new icon

        Parameters
        ----------
        widget : Gtk.Widget
            Object which receive signal
        """

        dialog = IconsDialog(self, _("Choose an icon"), self.path, "consoles")

        if dialog.new_path is not None:
            self.path = Path(dialog.new_path).expanduser()

            # Update icon thumbnail
            self.set_icon(self.image_console, self.path)

            # Update icon entry
            self.entry_icon.set_text(dialog.new_path)

        dialog.destroy()

    def __on_append_item(self, widget):
        """ Append a new row in treeview

        Parameters
        ----------
        widget : Gtk.Widget
            Object which receive signal
        """

        self.model_ignores.append([str()])

    def __on_remove_item(self, widget):
        """ Remove a row in treeview

        Parameters
        ----------
        widget : Gtk.Widget
            Object which receive signal
        """

        model, treeiter = self.treeview_ignores.get_selection().get_selected()
        if treeiter is not None:
            self.model_ignores.remove(treeiter)

    def __on_edited_cell(self, widget, path, text):
        """ Update treerow when a cell has been edited

        Parameters
        ----------
        widget : Gtk.Widget
            Object which receive signal
        path : str
            Path identifying the edited cell
        text : str
            New text
        """

        self.model_ignores[path][0] = str(text)

    def __on_check_advanced_mode(self, *args):
        """ Check advanced checkbutton status and update widgets sensitivity
        """

        status = self.check_advanced.get_active()

        self.label_recursive.set_visible(status)
        self.switch_recursive.set_visible(status)

        self.label_ignores.set_visible(status)
        self.grid_ignores.set_visible(status)
        self.scroll_ignores.set_visible(status)
        self.viewport_ignores.set_visible(status)
        self.treeview_ignores.set_visible(status)
        self.image_ignores_add.set_visible(status)
        self.button_ignores_add.set_visible(status)
        self.image_ignores_remove.set_visible(status)
        self.button_ignores_remove.set_visible(status)

    def set_icon(self, widget, path, size=64):
        """ Set thumbnail icon from a specific path

        Parameters
        ----------
        widget : Gtk.Widget
            Icon widget to update
        path : pathlib.Path
            Icon path
        size : int, optional
            Icon size in pixels (Default: 64)
        """

        # Retrieve an empty icon
        icon = self.main_parent.icons_blank.get(size)

        if path is not None:

            # Check icon from icons theme
            if not path.exists():
                collection_path = self.api.get_local("icons", f"{path}.png")

                # Retrieve icon from collection
                if collection_path.exists() and collection_path.is_file():

                    # Check the file mime-type to avoid non-image file
                    if magic_from_file(collection_path,
                                       mime=True).startswith("image/"):

                        icon = GdkPixbuf.Pixbuf.new_from_file_at_scale(
                            str(collection_path), size, size, True)

            # Retrieve icon from file
            elif path.is_file():

                # Check the file mime-type to avoid non-image file
                if magic_from_file(path, mime=True).startswith("image/"):
                    icon = GdkPixbuf.Pixbuf.new_from_file_at_scale(
                        str(path), size, size, True)

        widget.set_from_pixbuf(icon)



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# Filesystem
from pathlib import Path

# System
from configparser import ConfigParser


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class Configuration(ConfigParser):

    def __init__(self, filepath, **kwargs):
        """ Constructor

        Parameters
        ----------
        filepath : pathlib.Path
            Configuration file path

        Raises
        ------
        TypeError
            If filepath instance is not pathlib.Path
        """

        if not isinstance(filepath, Path):
            raise TypeError("Expected %s type for filepath, get %s" % (
                "pathlib.Path", type(filepath)))

        # ------------------------------------
        #   Variables
        # ------------------------------------

        ConfigParser.__init__(self, **kwargs)

        self.path = filepath.expanduser()

        self.reload()

    def __str__(self):
        """ Formated informations

        Returns
        -------
        str
            Formated string
        """

        text = list()

        for section in sorted(self.sections()):
            text.append("[%s]" % section)

            for key in sorted(self.options(section)):
                text.append("%s = %s" % (
                    key, self.get(section, key, fallback=str())))

        return '\n'.join(text)

    def reload(self):
        """ Read again the configuration file if exists
        """

        if self.path.exists():
            self.read(self.path)

    def item(self, section, option, default=None):
        """ Return an item from configuration

        Parameters
        ----------
        section : str
            Section name
        option : str
            Option name
        default : str, optional
            Fallback value to return if nothing has been founded
            (default: None)

        Returns
        -------
        str or None
            Value from section option
        """

        if self.has_section(section) and self.has_option(section, option):
            return self.get(section, option)

        return default

    def append(self, section, option, value):
        """ Append a new section to configuration

        Parameters
        ----------
        section : str
            Section name
        option : str
            Option name
        value : str
            Value to set

        Raises
        ------
        IndexError
            If the configuration file not has section
        """

        if self.has_section(section):
            raise IndexError("Cannot find %s section" % section)

        self.add_section(section)
        self.set(section, option, str(value))

    def modify(self, section, option, value):
        """ Modify a section from configuration

        This function append or update a section into configuration

        Parameters
        ----------
        section : str
            Section name
        option : str
            Option name
        value : object
            Value to set
        """

        # Set a specific value for bool variables
        if type(value) is bool:
            if value:
                value = "yes"
            else:
                value = "no"

        if self.has_section(section):
            self.set(section, option, str(value))

        else:
            self.append(section, option, str(value))

    def rename(self, section, new_name):
        """ Rename a section from configuration

        Parameters
        ----------
        section : str
            Previous section name
        new_name : str
            New section name
        """

        if self.has_section(section) and not self.has_section(new_name):

            for (option, value) in self.items(section):
                self.modify(new_name, option, value)

            self.remove(section)

    def remove(self, section):
        """ Remove a section from configuration

        Parameters
        ----------
        section : str
            Section name
        """

        if self.has_section(section):
            self.remove_section(section)

    def update(self):
        """ Write all data from cache into configuration file

        This function need to be exec after each modifications to update file
        content
        """

        with self.path.open('w') as pipe:
            self.write(pipe)

    def add_missing_data(self, secondary_path):
        """ Append to configuration all missing data from another configuration

        Parameters
        ----------
        secondary_path : str
            Configuration file where to get missing sections/options

        Raises
        ------
        OSError
            If the configuration filepath not exists
        """

        if not secondary_path.exists():
            raise OSError(2, "Cannot find file", str(secondary_path))

        config = Configuration(secondary_path)

        for section in config.sections():

            if not self.has_section(section):
                self.add_section(section)

            for option in config.options(section):

                if not self.has_option(section, option):
                    self.set(section, option, config.get(section, option))

        self.update()



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# Geode
from geode_gem.widgets.common import GeodeGtkCommon

# GObject
from gi.repository import Gtk


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class CommonMenuItem(GeodeGtkCommon):

    __setters__ = {
        "set_use_underline": True,
    }

    def __init__(self, subclass, title, *args, **kwargs):
        """ Constructor

        Parameters
        ----------
        subclass : Gtk.MenuItem
            Subclass widget type
        title : str
            Menu item label
        """

        GeodeGtkCommon.__init__(self, subclass, **kwargs)

        # Properties
        self.set_label(title)

        if args:
            submenu = GeodeGtkMenu(*args)

            self.append_widget(submenu)
            self.set_submenu(submenu)


class GeodeGtkMenu(GeodeGtkCommon, Gtk.Menu):

    def __init__(self, *args, **kwargs):
        """ Constructor
        """

        GeodeGtkCommon.__init__(self, Gtk.Menu, **kwargs)

        for child in args:
            self.append(child)

    def append(self, child):
        """ Append a new child at the end of menu list

        Parameters
        ----------
        child : Gtk.MenuItem
            New menu item to append
        """

        # Generate a dynamic name for Gtk.SeparatorMenuItem
        if child is None:
            child = Gtk.SeparatorMenuItem.new()

        elif isinstance(child, Gtk.RadioMenuItem) and child.group is not None:
            child.join_group(self.get_widget(child.group))

        self.append_widget(child)

        super().append(child)

    def remove(self, child):
        """ Remove a specific widget from menu list

        Parameters
        ----------
        child : Gtk.MenuItem
            Menu item to remove
        """

        if hasattr(child, "identifier") and self.has_widget(child.identifier):
            del self.inner_widgets[child.identifier]

        super().remove(child)

    def clear(self):
        """ Remove all items from menu list
        """

        for child in self.get_children():
            self.remove(child)


class GeodeGtkMenuItem(CommonMenuItem, Gtk.MenuItem):

    def __init__(self, *args, **kwargs):
        """ See geode_gem.ui.widgets.menu.CommonMenuItem
        """

        CommonMenuItem.__init__(self, Gtk.MenuItem, *args, **kwargs)


class GeodeGtkCheckMenuItem(CommonMenuItem, Gtk.CheckMenuItem):

    def __init__(self, *args, **kwargs):
        """ See geode_gem.ui.widgets.menu.CommonMenuItem
        """

        CommonMenuItem.__init__(self, Gtk.CheckMenuItem, *args, **kwargs)


class GeodeGtkRadioMenuItem(CommonMenuItem, Gtk.RadioMenuItem):

    def __init__(self, *args, **kwargs):
        """ See geode_gem.ui.widgets.menu.CommonMenuItem
        """

        CommonMenuItem.__init__(self, Gtk.RadioMenuItem, *args, **kwargs)

        self.group = kwargs.get("group", None)



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# Geode
from geode_gem.widgets.view import CommonView

# GObject
from gi.repository import Gtk


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class GeodeGtkIconView(CommonView, Gtk.IconView):

    __setters__ = {
        "set_column_spacing": 0,
        "set_has_tooltip": False,
        "set_item_width": 32,
        "set_row_spacing": 0,
        "set_selection_mode": Gtk.SelectionMode.SINGLE,
        "set_spacing": 0,
    }

    def __init__(self, model, **kwargs):
        """ Constructor

        Parameters
        ----------
        """

        CommonView.__init__(self, Gtk.IconView, model, **kwargs)

        # Settings
        self.set_model(self.inner_model)

    def get_selected_treeiter(self):
        """ See geode_gem.widgets.view.CommonView.get_selected_treeiter
        """

        items_list = self.get_selected_items()

        if items_list:
            return self.inner_model.get_iter(items_list[0])

        return None

    def select_path_and_scroll(self, treepath, **kwargs):
        """ See geode_gem.widgets.view.CommonView.select_path_and_scroll
        """

        self.select_path(treepath)

        self.scroll_to_path(treepath,
                            kwargs.get("use_align", True),
                            kwargs.get("row_align", 0.5),
                            kwargs.get("col_align", 0.5))



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# GEM
from geode_gem.ui.data import Icons
from geode_gem.ui.widgets.window import CommonWindow

# GObject
from gi.repository import Gtk, Pango

# Translation
from gettext import gettext as _


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class MednafenDialog(CommonWindow):

    def __init__(self, parent, name, data):
        """ Constructor

        Parameters
        ----------
        parent : Gtk.Window
            Parent object
        name : str
            Game name
        data : dict
            Backup memory type dictionary (with type as key)
        """

        classic_theme = False
        if parent is not None:
            classic_theme = parent.use_classic_theme

        CommonWindow.__init__(self,
                              parent,
                              _("Backup Memory Type"),
                              Icons.Symbolic.SAVE,
                              classic_theme)

        # ------------------------------------
        #   Initialize variables
        # ------------------------------------

        self.data = data

        self.name = name

        self.memory_list = ["eeprom", "flash", "rtc", "sensor", "sram"]

        # ------------------------------------
        #   Prepare interface
        # ------------------------------------

        # Init widgets
        self.__init_widgets()

        # Init signals
        self.__init_signals()

        # Start interface
        self.__start_interface()

    def __init_widgets(self):
        """ Initialize interface widgets
        """

        self.set_size(640, 420)

        # ------------------------------------
        #   Grid
        # ------------------------------------

        grid = Gtk.Grid()

        # Properties
        grid.set_row_spacing(6)
        grid.set_column_spacing(12)

        # ------------------------------------
        #   Description
        # ------------------------------------

        label = Gtk.Label()
        label_game = Gtk.Label()

        # Properties
        label.set_text(_("This dialog allows you to specify specific backup "
                         "memory type for the following game:"))
        label.set_line_wrap(True)
        label.set_max_width_chars(8)
        label.set_single_line_mode(False)
        label.set_justify(Gtk.Justification.FILL)
        label.set_line_wrap_mode(Pango.WrapMode.WORD)

        label_game.set_text(self.name)
        label_game.set_margin_bottom(12)
        label_game.set_single_line_mode(True)
        label_game.set_ellipsize(Pango.EllipsizeMode.END)
        label_game.get_style_context().add_class(Gtk.STYLE_CLASS_DIM_LABEL)

        # ------------------------------------
        #   Buttons
        # ------------------------------------

        buttons = Gtk.Box()

        self.image_add = Gtk.Image()
        self.button_add = Gtk.Button()

        self.image_remove = Gtk.Image()
        self.button_remove = Gtk.Button()

        # Properties
        Gtk.StyleContext.add_class(
            buttons.get_style_context(), "linked")
        buttons.set_spacing(-1)
        buttons.set_orientation(Gtk.Orientation.VERTICAL)

        self.image_add.set_from_icon_name(
            Icons.Symbolic.ADD, Gtk.IconSize.MENU)
        self.button_add.set_image(self.image_add)

        self.image_remove.set_from_icon_name(
            Icons.Symbolic.REMOVE, Gtk.IconSize.MENU)
        self.button_remove.set_image(self.image_remove)

        # ------------------------------------
        #   Link
        # ------------------------------------

        link = Gtk.LinkButton()

        # Properties
        link.set_label(_("Mednafen GBA documentation"))
        link.set_uri("https://mednafen.github.io/documentation/gba.html"
                     "#Section_backupmem_type")

        # ------------------------------------
        #   Content list
        # ------------------------------------

        frame = Gtk.Frame()
        scroll = Gtk.ScrolledWindow()
        view = Gtk.Viewport()

        self.adjustment_value = Gtk.Adjustment()

        self.model_memory_keys = Gtk.ListStore(str)

        self.model = Gtk.ListStore(str, int)
        self.treeview = Gtk.TreeView()

        column = Gtk.TreeViewColumn()

        self.cell_key = Gtk.CellRendererCombo()
        self.cell_value = Gtk.CellRendererSpin()

        # Properties
        self.adjustment_value.set_lower(0)
        self.adjustment_value.set_upper(2147483647)  # INT_MAX
        self.adjustment_value.set_step_increment(16)
        self.adjustment_value.set_page_increment(1024)

        self.treeview.set_hexpand(True)
        self.treeview.set_vexpand(True)
        self.treeview.set_model(self.model)
        self.treeview.set_headers_clickable(False)
        self.treeview.set_headers_visible(False)
        self.treeview.set_show_expanders(False)
        self.treeview.set_has_tooltip(False)

        column.set_expand(True)
        column.pack_start(self.cell_key, True)
        column.pack_start(self.cell_value, True)

        column.add_attribute(self.cell_key, "text", 0)
        column.add_attribute(self.cell_value, "text", 1)

        self.cell_key.set_padding(12, 6)
        self.cell_key.set_alignment(0, .5)
        self.cell_key.set_property("text-column", 0)
        self.cell_key.set_property("editable", True)
        self.cell_key.set_property("has-entry", False)
        self.cell_key.set_property("model", self.model_memory_keys)

        self.cell_value.set_padding(12, 6)
        self.cell_value.set_alignment(1, .5)
        self.cell_value.set_property("editable", True)
        self.cell_value.set_property("adjustment", self.adjustment_value)

        # ------------------------------------
        #   Integrate widgets
        # ------------------------------------

        frame.add(self.treeview)

        scroll.add(view)
        view.add(grid)

        self.treeview.append_column(column)

        buttons.add(self.button_add)
        buttons.add(self.button_remove)

        grid.attach(label, 0, 0, 2, 1)
        grid.attach(label_game, 0, 1, 2, 1)
        grid.attach(buttons, 0, 2, 1, 1)
        grid.attach(frame, 1, 2, 1, 1)
        grid.attach(link, 0, 3, 2, 1)

        self.pack_start(scroll, True, True)

    def __init_signals(self):
        """ Initialize widgets signals
        """

        self.button_add.connect("clicked", self.__on_append_item)
        self.button_remove.connect("clicked", self.__on_remove_item)

        self.cell_key.connect("edited", self.__on_edited_cell)
        self.cell_value.connect("edited", self.__on_edited_cell)

    def __start_interface(self):
        """ Load data and start interface
        """

        self.add_button(_("Cancel"), Gtk.ResponseType.CLOSE)
        self.add_button(_("Apply"), Gtk.ResponseType.APPLY, Gtk.Align.END)

        for key in self.memory_list:
            self.model_memory_keys.append([key])

        for key, value in self.data.items():
            self.model.append([key, value])

        self.show_all()

    def __on_append_item(self, widget):
        """ Append a new row in treeview

        Parameters
        ----------
        widget : Gtk.Widget
            Object which receive signal
        """

        self.model.append([self.memory_list[0], int()])

    def __on_remove_item(self, widget):
        """ Remove a row in treeview

        Parameters
        ----------
        widget : Gtk.Widget
            Object which receive signal
        """

        model, treeiter = self.treeview.get_selection().get_selected()
        if treeiter is not None:
            self.model.remove(treeiter)

    def __on_edited_cell(self, widget, path, text):
        """ Update treerow when a cell has been edited

        Parameters
        ----------
        widget : Gtk.Widget
            Object which receive signal
        path : str
            Path identifying the edited cell
        text : str
            New text
        """

        if type(widget) == Gtk.CellRendererCombo:
            self.model[path][0] = str(text)

        elif type(widget) == Gtk.CellRendererSpin:
            self.model[path][1] = int(text)



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# Datetime
import datetime

# Filesystem
from os import R_OK, W_OK, access

# Geode
from geode_gem.engine.utils import get_binary_path

# GObject
from gi.repository import GdkPixbuf, Gtk

# Processus
from subprocess import PIPE, Popen, STDOUT

# Regex
from re import match as re_match
from re import escape as re_escape
from re import findall as re_findall

# Translation
from gettext import gettext as _
from gettext import ngettext


# ------------------------------------------------------------------------------
#   Misc functions
# ------------------------------------------------------------------------------

def call_external_application(*command):
    """ Call an external application from operating system

    Parameters
    ----------
    command : list
        Command execution parameters as string list

    Returns
    -------
    str
        Splited output if executed, None otherwise
    """

    if get_binary_path(command[0]):
        proc = Popen(
            command,
            stdin=PIPE,
            stdout=PIPE,
            stderr=STDOUT,
            universal_newlines=True)

        output, error_output = proc.communicate()
        if output:
            return output.split('\n')[0]

    return None


def check_game_is_editable(game):
    """ Check if specified game is a text file and can be edited by user

    Parameters
    ----------
    game : geode_gem.engine.game.Game
        Game instance

    Returns
    -------
    bool
        True if game file is editable, False otherwise
    """

    if not magic_from_file(game.path, mime=True).startswith("text/"):
        return False

    return access(game.path, R_OK) and access(game.path, W_OK)


def check_mednafen():
    """ Check if Mednafen exists on user system

    This function read the first line of mednafen default output and check
    if this one match "Starting Mednafen [0-9+.?]+".

    Returns
    -------
    bool
        Mednafen exists status

    Notes
    -----
    Still possible to troll this function with a script call mednafen which
    send the match string as output. But, this problem only appear if a
    user want to do that, so ...
    """

    output = call_external_application("mednafen")

    return (output and
            re_match(r'Starting Mednafen [\d+\.?]+', output.split('\n')[0]))


def generate_blank_icons():
    """ Generate multiple blank icons to use as empty Pixbuf

    Returns
    -------
    dict
        Icons storage
    """

    storage = dict()

    for size in (8, 16, 22, 24, 32, 48, 64, 96):
        storage[size] = GdkPixbuf.Pixbuf.new(
            GdkPixbuf.Colorspace.RGB, True, 8, size, size)
        storage[size].fill(0x00000000)

    return storage


def get_today_from_datetime(date_object):
    """ Retrieve current time for a specific datetime type

    This function is only used to test string_from_date with unittest.mock

    Parameters
    ----------
    date_object : datetime.datetime
        Date to retrieve datetime type

    Returns
    -------
    datetime.date or datetime.datetime
        Current time
    """

    if isinstance(date_object, datetime.datetime):
        return datetime.datetime.now()

    return datetime.date.today()


def icon_from_data(path, fallback=None, width=24, height=24):
    """ Load an icon from path

    This function search if an icon is available in GEM icons folder and return
    a generate Pixbuf from the icon path. If no icon was found, return an empty
    generate Pixbuf.

    Parameters
    ----------
    path : pathlib.Path
        Absolute or relative icon path
    fallback : GdkPixbuf.Pixbuf, optional
        Fallback icon to return instead of empty (Default: None)
    width : int, optional
        Icon width in pixels (Default: 24)
    height : int, optional
        Icon height in pixels (Default: 24)

    Returns
    -------
    GdkPixbuf.Pixbuf
        Pixbuf icon object
    """

    if path.exists():

        try:
            return GdkPixbuf.Pixbuf.new_from_file_at_size(
                str(path), width, height)
        except Exception:
            pass

    # Return an empty icon
    if fallback is None:
        fallback = GdkPixbuf.Pixbuf.new(
            GdkPixbuf.Colorspace.RGB, True, 8, width, height)
        fallback.fill(0x00000000)

    return fallback


def on_activate_listboxrow(listbox, row):
    """ Activate internal widget when a row has been activated

    Parameters
    ----------
    listbox : Gtk.ListBox
        Object which receive signal
    row : gem.widgets.widgets.ListBoxItem
        Activated row
    """

    widget = row.get_widget()

    if type(widget) == Gtk.ComboBox:
        widget.popup()

    elif type(widget) == Gtk.Entry:
        widget.grab_focus()

    elif type(widget) == Gtk.FileChooserButton:
        widget.activate()

    elif type(widget) in [Gtk.Button, Gtk.FontButton]:
        widget.clicked()

    elif type(widget) == Gtk.SpinButton:
        widget.grab_focus()

    elif type(widget) == Gtk.Switch:
        widget.set_active(not widget.get_active())


def on_prefer_dark_theme(status=False):
    """ Change dark status of interface theme

    Parameters
    ----------
    status : bool, optional
        Use dark theme (Default: False)
    """

    settings = Gtk.Settings.get_default()
    if settings:
        settings.set_property("gtk-application-prefer-dark-theme", status)


def on_entry_clear(widget, pos, event):
    """ Reset an entry widget when secondary icon is clicked

    Parameters
    ----------
    widget : Gtk.Entry
        Entry widget
    pos : Gtk.EntryIconPosition
        Position of the clicked icon
    event : Gdk.EventButton or Gdk.EventKey
        Event which triggered this signal

    Returns
    -------
    bool
        Function state
    """

    if type(widget) is not Gtk.Entry:
        return False

    if pos == Gtk.EntryIconPosition.SECONDARY and widget.get_text():
        widget.set_text(str())

        return True

    return False


def magic_from_file(filename, mime=False):
    """ Fallback function to retrieve file type when python-magic is missing

    Parameters
    ----------
    filename : str
        File path to read
    mime : bool
        Retrieve the file mimetype, otherwise a readable name (Default: False)

    Returns
    -------
    str
        File type result as human readable name or mimetype
    """

    # Use dereference to follow symlink
    commands = ["file", "--dereference", str(filename)]
    if mime:
        commands.insert(1, "--mime-type")

    output = call_external_application(*commands)
    if output:
        output = output.splitlines()[0]

        result = re_findall(fr"^{re_escape(str(filename))}\:\s+(.*)$", output)
        if result:
            return result[0]

    return str()


def replace_for_markup(text):
    """ Replace some characters in text for markup compatibility

    Parameters
    ----------
    text : str
        Text to parser

    Returns
    -------
    str
        Replaced text
    """

    characters = {
        '&': "&amp;",
        '<': "&lt;",
        '>': "&gt;",
    }

    for key, value in characters.items():
        text = text.replace(key, value)

    return text


def string_from_date(date_object):
    """ Convert a datetime to a pretty string

    Get a pretty string from the interval between NOW() and the wanted date

    Parameters
    ----------
    date_object : datetime.datetime
        Date to compare with NOW()

    Returns
    -------
    str or None
        Convert value
    """

    if date_object is None:
        return None

    days = int((get_today_from_datetime(date_object) - date_object).days)

    if days == 0:
        return _("Today")
    elif days == 1:
        return _("Yesterday")
    elif days == -1:
        return _("Tomorrow")
    elif days > 0 and days < 30:
        return _("%d days ago") % abs(days)
    elif days < 0 and days > -30:
        return _("In %d days") % abs(days)

    months = int(days / 30)

    if months == 1:
        return _("Last month")
    elif months == -1:
        return _("Next month")
    elif months > 0 and months < 12:
        return _("%d months ago") % abs(months)
    elif months < 0 and months > -12:
        return _("In %d months") % abs(months)

    years = int(months / 12)

    if years == 1:
        return _("Last year")
    elif years == -1:
        return _("Next year")
    elif years > 1:
        return _("%d years ago") % abs(years)

    return _("In %d years") % abs(years)


def string_from_time(time_object):
    """ Convert a time to a pretty string

    Get a pretty string from the interval between NOW() and the wanted date

    Parameters
    ----------
    time_object : datetime.datetime
        Date to compare with NOW()

    Returns
    -------
    str or None
        Convert value
    """

    if time_object is None:
        return None

    hours, minutes, seconds = int(), int(), int()

    if type(time_object) is datetime.timedelta:
        hours, seconds = divmod(time_object.seconds, 3600)

        if seconds > 0:
            minutes, seconds = divmod(seconds, 60)

        hours += time_object.days * 24

    if hours == 0 and minutes == 0 and seconds == 0:
        return str()

    elif hours == 0 and minutes == 0:
        return ngettext(_("1 second"), _("%d seconds") % seconds, seconds)

    elif hours == 0:
        return ngettext(_("1 minute"), _("%d minutes") % minutes, minutes)

    return ngettext(_("1 hour"), _("%d hours") % hours, hours)



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

from geode_gem.ui.dialog.cache import CleanCacheDialog
from geode_gem.ui.dialog.cover import CoverDialog
from geode_gem.ui.dialog.delete import DeleteDialog
from geode_gem.ui.dialog.dndconsole import DNDConsoleDialog
from geode_gem.ui.dialog.duplicate import DuplicateDialog
from geode_gem.ui.dialog.editor import EditorDialog
from geode_gem.ui.dialog.maintenance import MaintenanceDialog
from geode_gem.ui.dialog.mednafen import MednafenDialog
from geode_gem.ui.dialog.message import MessageDialog
from geode_gem.ui.dialog.parameter import ParametersDialog
from geode_gem.ui.dialog.question import QuestionDialog
from geode_gem.ui.dialog.rename import RenameDialog
from geode_gem.ui.dialog.viewer import ViewerDialog
from geode_gem.ui.preferences.interface import (ConsolePreferences,
                                                EmulatorPreferences,
                                                PreferencesWindow)


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class GeodeDialog:
    """ Custom dialogs for Geode-GEM applications
    """

    CleanCache = CleanCacheDialog
    Console = ConsolePreferences
    Cover = CoverDialog
    DNDConsole = DNDConsoleDialog
    Delete = DeleteDialog
    Duplicate = DuplicateDialog
    Editor = EditorDialog
    Emulator = EmulatorPreferences
    Maintenance = MaintenanceDialog
    Mednafen = MednafenDialog
    Message = MessageDialog
    Parameters = ParametersDialog
    Preferences = PreferencesWindow
    Question = QuestionDialog
    Rename = RenameDialog
    Viewer = ViewerDialog



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# GEM
from geode_gem.ui.data import Icons
from geode_gem.ui.utils import replace_for_markup
from geode_gem.ui.widgets.window import CommonWindow

# GObject
from gi.repository import GdkPixbuf, Gtk, Pango

# Translation
from gettext import gettext as _


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class CoverDialog(CommonWindow):

    def __init__(self, parent, game):
        """ Constructor

        Parameters
        ----------
        parent : Gtk.Window
            Parent object
        game : gem.api.Game
            Game object instance
        """

        classic_theme = False
        if parent is not None:
            classic_theme = parent.use_classic_theme

        CommonWindow.__init__(
            self, parent, _("Game cover"), Icons.Symbolic.IMAGE, classic_theme)

        # ------------------------------------
        #   Initialize variables
        # ------------------------------------

        self.game = game

        self.mimetypes = [
            "image/bmp",
            "image/jpeg",
            "image/png",
            "image/svg+xml",
            "image/svg+xml-compressed",
        ]

        # ------------------------------------
        #   Prepare interface
        # ------------------------------------

        # Init widgets
        self.__init_widgets()

        # Init signals
        self.__init_signals()

        # Start interface
        self.__start_interface()

    def __init_widgets(self):
        """ Initialize interface widgets
        """

        self.set_size(640, 480)

        self.set_spacing(6)

        self.set_resizable(True)

        # ------------------------------------
        #   Grid
        # ------------------------------------

        self.grid_content = Gtk.Box()

        # Properties
        self.grid_content.set_spacing(6)

        # ------------------------------------
        #   Title
        # ------------------------------------

        self.label_title = Gtk.Label()

        # Properties
        self.label_title.set_markup(
            "<span weight='bold' size='large'>%s</span>" % (
                replace_for_markup(self.game.name)))
        self.label_title.set_use_markup(True)
        self.label_title.set_halign(Gtk.Align.CENTER)
        self.label_title.set_ellipsize(Pango.EllipsizeMode.END)

        # ------------------------------------
        #   Image selector
        # ------------------------------------

        self.label_image_selector = Gtk.Label()

        self.filter_image_selector = Gtk.FileFilter.new()

        self.dialog_image_selector = Gtk.FileChooserDialog(
            use_header_bar=not self.use_classic_theme)

        self.file_image_selector = Gtk.FileChooserButton.new_with_dialog(
            self.dialog_image_selector)

        self.image_reset = Gtk.Image()
        self.button_reset = Gtk.Button()

        # Properties
        self.label_image_selector.set_markup("<b>%s</b>" % _("Thumbnail file"))
        self.label_image_selector.set_margin_top(12)
        self.label_image_selector.set_hexpand(True)
        self.label_image_selector.set_use_markup(True)
        self.label_image_selector.set_single_line_mode(True)
        self.label_image_selector.set_halign(Gtk.Align.CENTER)
        self.label_image_selector.set_ellipsize(Pango.EllipsizeMode.END)

        for pattern in self.mimetypes:
            self.filter_image_selector.add_mime_type(pattern)

        self.dialog_image_selector.add_button(
            _("Cancel"), Gtk.ResponseType.CANCEL)
        self.dialog_image_selector.add_button(
            _("Accept"), Gtk.ResponseType.ACCEPT)
        self.dialog_image_selector.set_filter(self.filter_image_selector)
        self.dialog_image_selector.set_action(Gtk.FileChooserAction.OPEN)
        self.dialog_image_selector.set_create_folders(False)
        self.dialog_image_selector.set_local_only(True)

        self.file_image_selector.set_hexpand(True)

        self.image_reset.set_from_icon_name(
            Icons.Symbolic.CLEAR, Gtk.IconSize.BUTTON)

        # ------------------------------------
        #   Image preview
        # ------------------------------------

        self.label_preview = Gtk.Label()

        self.scroll_preview = Gtk.ScrolledWindow()
        self.view_preview = Gtk.Viewport()

        self.image_preview = Gtk.Image()

        # Properties
        self.label_preview.set_markup("<b>%s</b>" % _("Preview"))
        self.label_preview.set_margin_top(12)
        self.label_preview.set_hexpand(True)
        self.label_preview.set_use_markup(True)
        self.label_preview.set_single_line_mode(True)
        self.label_preview.set_halign(Gtk.Align.CENTER)
        self.label_preview.set_ellipsize(Pango.EllipsizeMode.END)

        self.scroll_preview.set_shadow_type(Gtk.ShadowType.OUT)

        self.image_preview.set_halign(Gtk.Align.CENTER)
        self.image_preview.set_valign(Gtk.Align.CENTER)
        self.image_preview.set_hexpand(True)
        self.image_preview.set_vexpand(True)

        # ------------------------------------
        #   Integrate widgets
        # ------------------------------------

        self.button_reset.add(self.image_reset)

        self.grid_content.pack_start(self.file_image_selector, True, True, 0)
        self.grid_content.pack_start(self.button_reset, False, False, 0)

        self.view_preview.add(self.image_preview)
        self.scroll_preview.add(self.view_preview)

        self.pack_start(self.label_title, False, False)
        self.pack_start(self.label_image_selector, False, False)
        self.pack_start(self.grid_content, False, False)
        self.pack_start(self.label_preview, False, False)
        self.pack_start(self.scroll_preview)

    def __init_signals(self):
        """ Initialize widgets signals
        """

        self.file_image_selector.connect("file-set", self.__update_preview)

        self.button_reset.connect("clicked", self.__on_reset_cover)

    def __start_interface(self):
        """ Load data and start interface
        """

        self.add_button(_("Cancel"), Gtk.ResponseType.CANCEL)
        self.add_button(_("Accept"), Gtk.ResponseType.APPLY, Gtk.Align.END)

        if self.game.cover is not None and self.game.cover.exists():
            self.file_image_selector.set_filename(str(self.game.cover))

        self.__update_preview()

    def __update_preview(self, *args):
        """ Update image preview
        """

        self.__on_set_preview(self.file_image_selector.get_filename())

    def __on_set_preview(self, path):
        """ Set a new preview from selector filepath

        Parameters
        ----------
        path : str
            Image file path
        """

        try:
            self.image_preview.set_from_pixbuf(
                GdkPixbuf.Pixbuf.new_from_file_at_scale(path, 400, 236, True))

        except Exception:
            self.__on_reset_cover()

    def __on_reset_cover(self, *args):
        """ Reset cover filechooser
        """

        self.file_image_selector.unselect_all()

        self.image_preview.set_from_icon_name(Icons.MISSING, Gtk.IconSize.DND)



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# Geode
from geode_gem.widgets.common import GeodeGtkCommon
from geode_gem.widgets.misc import GeodeGtkLabel

# GObject
from gi.repository import Gtk


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class GeodeGtkStackSidebar(GeodeGtkCommon, Gtk.Box):

    def __init__(self, *args, **kwargs):
        """ Constructor
        """

        GeodeGtkCommon.__init__(self, Gtk.StackSidebar, **kwargs)

        self.sidebar = Gtk.StackSidebar.new()

        self.stack = kwargs.get("use_stack", Gtk.Stack())
        self.stack.set_transition_type(Gtk.StackTransitionType.NONE)

        self.sidebar.set_stack(self.stack)

        for element in args:
            self.append_widget(element)
            self.stack.add_titled(element, element.identifier, element.title)

        self.pack_start(self.sidebar, False, False, 0)
        self.pack_start(self.stack, True, True, 0)


class GeodeGtkStackView(GeodeGtkCommon, Gtk.ScrolledWindow):

    def __init__(self, title, *args, **kwargs):
        """ Constructor
        """

        GeodeGtkCommon.__init__(self, Gtk.ScrolledWindow, **kwargs)

        self.inner_viewport = Gtk.Viewport.new(None, None)
        self.inner_grid = Gtk.Box.new(Gtk.Orientation.VERTICAL, 0)

        self.title = title

        # ------------------------------------
        #   Properties
        # ------------------------------------

        self.inner_grid.set_border_width(16)

        # ------------------------------------
        #   Packing
        # ------------------------------------

        for element in args:
            self.append_widget(element)
            self.inner_grid.pack_start(
                element,
                getattr(element, "is_fillable", False),
                getattr(element, "is_expandable", False),
                0)

        self.add(self.inner_viewport)
        self.inner_viewport.add(self.inner_grid)


class GeodeGtkStackSection(GeodeGtkCommon, Gtk.Box):

    __setters__ = {
        "set_orientation": Gtk.Orientation.VERTICAL,
        "set_homogeneous": False,
        "set_margin_bottom": 12,
        "set_spacing": 6,
    }

    def __init__(self, title, *args, **kwargs):
        """ Constructor
        """

        GeodeGtkCommon.__init__(self, Gtk.Box, **kwargs)

        label_title = GeodeGtkLabel(identifier="title",
                                    set_markup=f"<b>{title}</b>",
                                    set_style=Gtk.STYLE_CLASS_DIM_LABEL)

        # Packing
        self.pack_start(label_title, False, False, 0)

        for element in args:
            self.append_widget(element)
            self.pack_start(element, True, True, 0)


class GeodeGtkStackOption(GeodeGtkCommon, Gtk.Box):

    __setters__ = {
        "set_orientation": Gtk.Orientation.HORIZONTAL,
        "set_homogeneous": False,
        "set_spacing": 12,
    }

    def __init__(self, label, *args, **kwargs):
        """ Constructor
        """

        GeodeGtkCommon.__init__(self, Gtk.Box, **kwargs)

        label_option = GeodeGtkLabel(identifier="label",
                                     set_alignment=(1, 0.5),
                                     set_style=Gtk.STYLE_CLASS_DIM_LABEL,
                                     set_text=f"{label}")

        # Packing
        self.pack_start(label_option, False, False, 0)

        for element in args:
            self.append_widget(element)
            self.pack_start(element, True, True, 0)



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# Filesystem
from pathlib import Path

# GEM
from geode_gem.ui.data import Icons
from geode_gem.ui.utils import magic_from_file
from geode_gem.ui.widgets.window import CommonWindow

# GObject
from gi.repository import GdkPixbuf, GLib, Gtk

# Translation
from gettext import gettext as _


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class IconsDialog(CommonWindow):

    def __init__(self, parent, title, path, folder):
        """ Constructor

        Parameters
        ----------
        parent : Gtk.Window
            Parent object
        title : str
            Window title
        path : str
            Selected icon path
        folder : str
            Icons folder
        """

        CommonWindow.__init__(self,
                              parent,
                              title,
                              Icons.Symbolic.IMAGE,
                              parent.use_classic_theme)

        # ------------------------------------
        #   Variables
        # ------------------------------------

        self.api = parent.api

        # Current file path
        self.path = path
        # The new file path
        self.new_path = None

        # File selector management
        self.__file_path = None
        self.__file_active = False

        # Collection type
        self.folder = folder

        # Collection thread loading instance
        self.thread = int()

        # ------------------------------------
        #   Initialization
        # ------------------------------------

        # Init widgets
        self.__init_widgets()

        # Init packing
        self.__init_packing()

        # Init signals
        self.__init_signals()

        # Start interface
        self.__start_interface()

    def __init_widgets(self):
        """ Initialize interface widgets
        """

        self.set_size(800, 600)

        self.set_resizable(True)

        self.set_spacing(6)
        self.set_border_width(6)

        # ------------------------------------
        #   Grid
        # ------------------------------------

        self.stack = Gtk.Stack()
        self.stack_switcher = Gtk.StackSwitcher()

        # Properties
        self.stack.set_transition_type(Gtk.StackTransitionType.NONE)

        self.stack_switcher.set_stack(self.stack)
        self.stack_switcher.set_halign(Gtk.Align.CENTER)

        # ------------------------------------
        #   File patterns
        # ------------------------------------

        self.__file_patterns = Gtk.FileFilter.new()

        # ------------------------------------
        #   Custom
        # ------------------------------------

        self.frame_icons = Gtk.Frame()

        self.file_icons = Gtk.FileChooserWidget.new(Gtk.FileChooserAction.OPEN)

        # Properties
        self.frame_icons.set_shadow_type(Gtk.ShadowType.OUT)

        self.file_icons.set_hexpand(True)
        self.file_icons.set_vexpand(True)
        self.file_icons.set_filter(self.__file_patterns)
        self.file_icons.set_current_folder(str(Path.home()))

        # ------------------------------------
        #   Icons
        # ------------------------------------

        self.view_icons = Gtk.IconView()
        self.model_icons = Gtk.ListStore(GdkPixbuf.Pixbuf, str)

        self.scroll_icons = Gtk.ScrolledWindow()

        # Properties
        self.view_icons.set_model(self.model_icons)
        self.view_icons.set_pixbuf_column(0)
        self.view_icons.set_tooltip_column(1)
        self.view_icons.set_selection_mode(Gtk.SelectionMode.SINGLE)

        self.model_icons.set_sort_column_id(1, Gtk.SortType.ASCENDING)

        self.scroll_icons.set_hexpand(True)
        self.scroll_icons.set_vexpand(True)
        self.scroll_icons.set_shadow_type(Gtk.ShadowType.OUT)
        self.scroll_icons.set_policy(
            Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)

    def __init_packing(self):
        """ Initialize widgets packing in main window
        """

        if self.folder == "consoles":
            self.stack.add_titled(self.scroll_icons, "library", _("Library"))

            self.scroll_icons.add(self.view_icons)

            self.stack.add_titled(self.frame_icons, "file", _("File"))

            self.frame_icons.add(self.file_icons)

            self.pack_start(self.stack_switcher, False, False)
            self.pack_start(self.stack, True, True)

        else:
            self.frame_icons.add(self.file_icons)

            self.pack_start(self.frame_icons, True, True)

    def __init_signals(self):
        """ Initialize widgets signals
        """

        # Only activate the last child cause there are RadioButton
        children = self.stack_switcher.get_children()
        if len(children) > 0:
            children[-1].connect("clicked", self.__on_switch_stack_view)

        self.file_icons.connect("file-activated", self.__on_selected_icon)

        self.view_icons.connect("item-activated", self.__on_selected_icon)

    def __start_interface(self):
        """ Load data and start interface
        """

        self.add_button(_("Close"), Gtk.ResponseType.CLOSE)
        self.add_button(_("Accept"), Gtk.ResponseType.APPLY, Gtk.Align.END)

        self.load_interface()

        response = self.run()

        # Remove icons listing thread
        if not self.thread == 0:
            GLib.source_remove(self.thread)

        if response == Gtk.ResponseType.APPLY:
            self.save_interface()

    def __on_selected_icon(self, widget, path=None):
        """ Select an icon in treeview

        Parameters
        ----------
        widget : Gtk.Widget
            Object which receive signal
        path : Gtk.TreePath, optional
            Path to be activated (Default: None)
        """

        self.emit_response(None, Gtk.ResponseType.APPLY)

    def __on_switch_stack_view(self, widget):
        """ Check file selector to avoid to lost file selection

        Parameters
        ----------
        widget : Gtk.Widget
            Object which receive signal
        """

        # The focus is not in the file selector anymore
        if not widget.get_active():
            self.__file_active = False

            self.__file_path = self.file_icons.get_filename()

        elif not self.__file_active and self.__file_path is not None:
            self.__file_active = True

            self.file_icons.set_filename(self.__file_path)

    def load_interface(self):
        """ Insert data into interface's widgets
        """

        # Set authorized mime types for file selector
        self.__file_patterns.add_mime_type("image/*")

        # Fill icons view
        if self.api is not None:

            # Load icons collection
            if self.folder == "consoles":
                self.icons_data = dict()

                if not self.thread == 0:
                    GLib.source_remove(self.thread)

                self.thread = GLib.idle_add(self.append_icons(64).__next__)

            # Check the choosen path
            if self.path is not None and self.path.exists():

                if self.folder == "consoles":
                    self.frame_icons.show()
                    self.stack.set_visible_child(self.frame_icons)

                self.file_icons.set_filename(str(self.path))

                self.__file_path = self.path
                self.__file_active = True

    def save_interface(self):
        """ Return all the data from interface
        """

        # Retrieve path from file selector
        self.new_path = self.file_icons.get_filename()

        # Check icons collection
        if self.folder == "consoles":

            # Retrieve icon from icons collection
            if self.stack.get_visible_child_name() == "library":
                self.new_path = None

                items = self.view_icons.get_selected_items()

                # An icon is selected
                if len(items) > 0:
                    self.new_path = self.model_icons.get_value(
                        self.model_icons.get_iter(items[0]), 1)

    def append_icons(self, size):
        """ Append icons in icons view with a specific size

        Parameters
        ----------
        size : int
            Specified icon size
        """

        yield True

        # Retrieve files from icons collection
        collection_path = self.api.get_local("icons")

        for path in sorted(collection_path.glob("*.png")):

            # Retrieve an empty icon
            icon = self.main_parent.icons_blank.get(size)

            # Generate an icon for found file
            if path.exists() and path.is_file():

                # Check the file mime-type to avoid non-image file
                if magic_from_file(path, mime=True).startswith("image/"):
                    icon = GdkPixbuf.Pixbuf.new_from_file_at_scale(
                        str(path), size, size, True)

            self.icons_data[path.stem] = \
                self.model_icons.append([icon, path.stem])

            # Current icon match the choosen one
            if str(self.path) == path.stem:

                # Only select current icon if no selection is available
                if len(self.view_icons.get_selected_items()) == 0:
                    self.view_icons.select_path(
                        self.model_icons.get_path(self.icons_data[path.stem]))

            yield True

        # Remove thread id from memory
        self.thread = int()

        yield False



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# Geode
from geode_gem.widgets.misc import GeodeGtkBox

# GObject
from gi.repository import Gtk


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class GeodeGtkToolbar(GeodeGtkBox):

    __setters__ = {
        "set_spacing": 0,
        "set_border_width": 0,
        "set_orientation": Gtk.Orientation.HORIZONTAL,
    }

    def __init__(self, *args, **kwargs):
        """ Constructor
        """

        GeodeGtkBox.__init__(self, **kwargs)

        # ------------------------------------
        #   Packing
        # ------------------------------------

        for index, widget in enumerate(args):
            expand = False

            if widget is None:
                widget = Gtk.SeparatorToolItem.new()
                widget.set_expand(True)
                widget.set_draw(False)

                setattr(widget,
                        "identifier",
                        f"{self.identifier}_separator_{index}")

                expand = True

            elif len(args) == 1:
                expand = True

            self.append_widget(widget)
            self.pack_start(widget, expand, expand, 0)


class GeodeGtkToolbarBox(GeodeGtkBox):

    __setters__ = {
        "set_spacing": 0,
        "set_border_width": 0,
        "set_orientation": Gtk.Orientation.HORIZONTAL,
    }

    def __init__(self, *args, **kwargs):
        """ Constructor
        """

        GeodeGtkBox.__init__(self, **kwargs)

        # ------------------------------------
        #   Packing
        # ------------------------------------

        method = "get_hexpand"
        if self.get_orientation() == Gtk.Orientation.VERTICAL:
            method = "get_vexpand"

        for widget in args:
            expand = False
            if hasattr(widget, method):
                expand = getattr(widget, method)()

            self.append_widget(widget)
            self.pack_start(widget, expand, expand, 0)


class GeodeGtkToolbarSwitch(GeodeGtkToolbarBox):

    def __init__(self, *args, **kwargs):
        """ Constructor
        """

        GeodeGtkToolbarBox.__init__(self, *args, **kwargs)

        # Current toggled button
        self.toggled_identifier = kwargs.get("default", None)

        if self.has_widget(self.toggled_identifier):
            self.get_widget(self.toggled_identifier).set_active(True)

    def switch_to(self, widget_key):
        """ Toggled a specific widget

        Parameters
        ----------
        widget_key : str
            Internal widget keys, contains in self.inner_widgets
        """

        widget_activated = self.get_widget(widget_key)

        # Update buttons state
        for widget in self.inner_widgets.values():
            if isinstance(widget, Gtk.Button):
                widget.set_active(widget == widget_activated)

                # Store current toggled widget identifier
                self.toggled_identifier = widget.identifier

    def get_toggled_widget(self):
        """ Retrieve current toggled widget

        Returns
        -------
        Gtk.Button or None
            Toggled widget if founded, None otherwise
        """

        if self.has_widget(self.toggled_identifier):
            return self.get_widget(self.toggled_identifier)

        return None



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# GEM
from geode_gem.ui.data import Icons
from geode_gem.ui.utils import replace_for_markup
from geode_gem.ui.widgets.window import CommonWindow
from geode_gem.ui.widgets.widgets import ListBoxItem

# GObject
from gi.repository import Gtk, Pango

# Translation
from gettext import gettext as _


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class DeleteDialog(CommonWindow):

    def __init__(self, parent, game):
        """ Constructor

        Parameters
        ----------
        parent : Gtk.Window
            Parent object
        game : gem.engine.game.Game
            Game object
        """

        classic_theme = False
        if parent is not None:
            classic_theme = parent.use_classic_theme

        CommonWindow.__init__(self,
                              parent,
                              _("Remove a game"),
                              Icons.Symbolic.DELETE,
                              classic_theme)

        # ------------------------------------
        #   Variables
        # ------------------------------------

        self.game = game

        # ------------------------------------
        #   Prepare interface
        # ------------------------------------

        # Init widgets
        self.__init_widgets()

        # Init signals
        self.__init_signals()

        # Start interface
        self.__start_interface()

    def __init_widgets(self):
        """ Initialize interface widgets
        """

        self.set_size(640, 480)

        self.set_spacing(6)

        self.set_resizable(True)

        # ------------------------------------
        #   Grid
        # ------------------------------------

        self.grid_switch = Gtk.Grid()

        # Properties
        self.grid_switch.set_column_spacing(12)
        self.grid_switch.set_row_spacing(6)

        # ------------------------------------
        #   Title
        # ------------------------------------

        self.label_title = Gtk.Label()

        # Properties
        self.label_title.set_markup(
            "<span weight='bold' size='large'>%s</span>" % (
                replace_for_markup(self.game.name)))
        self.label_title.set_use_markup(True)
        self.label_title.set_halign(Gtk.Align.CENTER)
        self.label_title.set_ellipsize(Pango.EllipsizeMode.END)

        # ------------------------------------
        #   Description
        # ------------------------------------

        self.label_description = Gtk.Label()

        # Properties
        self.label_description.set_text(
            _("The following game going to be removed from your hard drive. "
              "This action is irreversible!"))
        self.label_description.set_line_wrap(True)
        self.label_description.set_max_width_chars(8)
        self.label_description.set_single_line_mode(False)
        self.label_description.set_justify(Gtk.Justification.FILL)
        self.label_description.set_line_wrap_mode(Pango.WrapMode.WORD)

        # ------------------------------------
        #   Options
        # ------------------------------------

        self.label_data = Gtk.Label()

        self.frame_options = Gtk.Frame()
        self.scroll_options = Gtk.ScrolledWindow()
        self.listbox_options = Gtk.ListBox()

        self.widget_database = ListBoxItem()
        self.switch_database = Gtk.Switch()

        self.widget_desktop = ListBoxItem()
        self.switch_desktop = Gtk.Switch()

        self.widget_savestate = ListBoxItem()
        self.switch_savestate = Gtk.Switch()

        self.widget_screenshots = ListBoxItem()
        self.switch_screenshots = Gtk.Switch()

        self.widget_cache = ListBoxItem()
        self.switch_cache = Gtk.Switch()

        self.widget_memory = ListBoxItem()
        self.switch_memory = Gtk.Switch()

        # Properties
        self.label_data.set_markup(
            "<b>%s</b>" % _("Optional data to remove"))
        self.label_data.set_margin_top(12)
        self.label_data.set_hexpand(True)
        self.label_data.set_use_markup(True)
        self.label_data.set_single_line_mode(True)
        self.label_data.set_halign(Gtk.Align.CENTER)
        self.label_data.set_ellipsize(Pango.EllipsizeMode.END)

        self.listbox_options.set_activate_on_single_click(True)
        self.listbox_options.set_selection_mode(
            Gtk.SelectionMode.NONE)

        self.widget_database.set_widget(self.switch_database)
        self.widget_database.set_option_label(
            _("Database"))
        self.widget_database.set_description_label(
            _("Delete game data from database"))

        self.widget_desktop.set_widget(self.switch_desktop)
        self.widget_desktop.set_option_label(
            _("Menu entry"))
        self.widget_desktop.set_description_label(
            _("Delete desktop file"))

        self.widget_savestate.set_widget(self.switch_savestate)
        self.widget_savestate.set_option_label(
            _("Savestates"))
        self.widget_savestate.set_description_label(
            _("Delete savestates files"))

        self.widget_screenshots.set_widget(self.switch_screenshots)
        self.widget_screenshots.set_option_label(
            _("Screenshots"))
        self.widget_screenshots.set_description_label(
            _("Delete screenshots files"))

        self.widget_cache.set_widget(self.switch_cache)
        self.widget_cache.set_option_label(
            _("Icons cache"))
        self.widget_cache.set_description_label(
            _("Delete generated icons from cache"))

        self.widget_memory.set_widget(self.switch_memory)
        self.widget_memory.set_option_label(
            _("Flash memory"))
        self.widget_memory.set_description_label(
            _("Delete flash memory file"))

        # ------------------------------------
        #   Integrate widgets
        # ------------------------------------

        self.listbox_options.add(self.widget_database)
        self.listbox_options.add(self.widget_desktop)
        self.listbox_options.add(self.widget_cache)
        self.listbox_options.add(self.widget_savestate)
        self.listbox_options.add(self.widget_screenshots)

        if self.main_parent.check_gba_game_use_mednafen(self.game):
            self.listbox_options.add(self.widget_memory)

        self.scroll_options.add(self.listbox_options)

        self.frame_options.add(self.scroll_options)

        self.pack_start(self.label_title, False, False)
        self.pack_start(self.label_description, False, False)
        self.pack_start(self.label_data, False, False)
        self.pack_start(self.frame_options)

    def __init_signals(self):
        """ Initialize widgets signals
        """

        self.listbox_options.connect(
            "row-activated", self.on_activate_listboxrow)

    def __start_interface(self):
        """ Load data and start interface
        """

        self.add_button(_("No"), Gtk.ResponseType.NO)
        self.add_button(_("Yes"), Gtk.ResponseType.YES, Gtk.Align.END)

        self.switch_database.set_active(True)
        self.switch_desktop.set_active(True)
        self.switch_cache.set_active(True)

    def get_data(self):
        """ Retrieve data to remove from user choices

        Returns
        -------
        dict
            Data to remove
        """

        data = {
            "paths": list(),
            "database": False
        }

        # ------------------------------------
        #   Game file
        # ------------------------------------

        data["paths"].append(self.game.path)

        # ------------------------------------
        #   Savestates
        # ------------------------------------

        if self.switch_savestate.get_active():
            data["paths"].extend(self.game.savestates)

        # ------------------------------------
        #   Screenshots
        # ------------------------------------

        if self.switch_screenshots.get_active():
            data["paths"].extend(self.game.screenshots)

        # ------------------------------------
        #   Desktop
        # ------------------------------------

        if self.switch_desktop.get_active():
            desktop_file = self.main_parent.get_game_desktop_file(self.game)
            if desktop_file.exists():
                data["paths"].append(desktop_file)

        # ------------------------------------
        #   Cache
        # ------------------------------------

        if self.switch_cache.get_active():

            for size in ("22x22", "96x96"):
                path = self.main_parent.get_icon_from_cache(
                    "games", size, "%s.png" % self.game.id)

                if path.exists():
                    data["paths"].append(path)

        # ------------------------------------
        #   Memory type
        # ------------------------------------

        if self.switch_memory.get_active():
            path = self.main_parent.get_mednafen_memory_type_file(self.game)

            if path.exists():
                data["paths"].append(path)

        # ------------------------------------
        #   Database
        # ------------------------------------

        if self.switch_database.get_active():
            data["database"] = True

        return data



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# GEM
from geode_gem.ui.data import Icons
from geode_gem.ui.widgets.window import CommonWindow

# GObject
from gi.repository import Gtk, Pango

# Translation
from gettext import gettext as _


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class QuestionDialog(CommonWindow):

    def __init__(self, parent, title, message):
        """ Constructor

        Parameters
        ----------
        parent : Gtk.Window
            Parent object
        title : str
            Dialog title
        message : str
            Dialog message
        """

        classic_theme = False
        if parent is not None:
            classic_theme = parent.use_classic_theme

        CommonWindow.__init__(
            self, parent, title, Icons.Symbolic.QUESTION, classic_theme)

        # ------------------------------------
        #   Variables
        # ------------------------------------

        self.message = message

        # ------------------------------------
        #   Prepare interface
        # ------------------------------------

        # Init widgets
        self.__init_widgets()

        # Start interface
        self.__start_interface()

    def __init_widgets(self):
        """ Initialize interface widgets
        """

        self.set_size(400, -1)

        # ------------------------------------
        #   Label
        # ------------------------------------

        self.label = Gtk.Label()

        # Properties
        self.label.set_line_wrap(True)
        self.label.set_use_markup(True)
        self.label.set_max_width_chars(10)
        self.label.set_line_wrap_mode(Pango.WrapMode.WORD)

        # ------------------------------------
        #   Integrate widgets
        # ------------------------------------

        self.pack_start(self.label, False, True)

    def __start_interface(self):
        """ Load data and start interface
        """

        self.add_button(_("No"), Gtk.ResponseType.NO)
        self.add_button(_("Yes"), Gtk.ResponseType.YES, Gtk.Align.END)

        self.label.set_markup(self.message)



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# Geode
from geode_gem.widgets.common import GeodeGtkCommon

# GObject
from gi.repository import Gtk


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class GeodeGtkStatusbar(GeodeGtkCommon, Gtk.Statusbar):

    pixbuf_widgets = list()

    def __init__(self, *args, **kwargs):
        """ Constructor
        """

        GeodeGtkCommon.__init__(self, Gtk.Statusbar, **kwargs)

        self.inner_grid = self.get_message_area()

        # Ensure to remove the default label widget from legacy Gtk.Statusbar
        label = self.inner_grid.get_children()[0]
        self.inner_grid.remove(label)
        label.destroy()

        for index, widget in enumerate(args):
            expand = False

            if widget is None:
                widget = Gtk.SeparatorToolItem.new()
                widget.set_expand(True)
                widget.set_draw(False)

                setattr(widget,
                        "identifier",
                        f"{self.identifier}_separator_{index}")

                expand = True

            elif isinstance(widget, Gtk.Label):
                widget.set_halign(Gtk.Align.START)

            elif isinstance(widget, Gtk.Image):
                self.pixbuf_widgets.append(widget.identifier)

            self.append_widget(widget)
            self.inner_grid.pack_start(widget, expand, expand, 0)

        # ------------------------------------
        #   Properties
        # ------------------------------------

        self.inner_grid.set_spacing(kwargs.get("spacing", 12))
        self.inner_grid.set_margin_top(kwargs.get("margin", 0))
        self.inner_grid.set_margin_end(kwargs.get("margin", 0))
        self.inner_grid.set_margin_start(kwargs.get("margin", 0))
        self.inner_grid.set_margin_bottom(kwargs.get("margin", 0))

    def set_widget_value(self, widget_key, **kwargs):
        """ Set an internal widget value

        The new value is set with the kwargs structure and based on widget type

        Gtk.Label → markup, text
        Gtk.Image → image, toolip
        Gtk.ProgressBar → index, length

        Parameters
        ----------
        widget_key : str
            Internal widget keys, contains in self.__dict__
        """

        widget = self.get_widget(widget_key)

        if isinstance(widget, Gtk.Label):
            text = kwargs.get("text", str()).strip()

            if widget.get_use_markup():
                widget.set_markup(text)
            else:
                widget.set_text(text)

        elif isinstance(widget, Gtk.Image):
            widget.set_tooltip_text(kwargs.get("tooltip", str()).strip())

            value = kwargs.get("image", None)
            if isinstance(value, tuple) and value:
                widget.set_from_icon_name(*value)
                value = value[0]
            else:
                widget.set_from_pixbuf(value)

            if value is None:
                widget.hide()
            else:
                widget.show()

        elif type(widget) is Gtk.ProgressBar:
            index = kwargs.get("index", int())
            length = kwargs.get("length", int())

            if length > 0:
                widget.set_text(f"{index}/{length}")
                widget.set_fraction(index / length)



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# Database
import sqlite3

# Filesystem
from pathlib import Path

# GEM
from geode_gem.engine.lib.configuration import Configuration

# Logging
from logging import Logger


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class Database(object):

    def __init__(self, db_path, configuration, logger):
        """ Constructor

        Parameters
        ----------
        db_path : pathlib.Path
            Database filepath
        configuration : gem.engine.lib.configuration.Configuration
            Configuration file which contains database scheme
        logger : logging.Logger
            Logging object

        Raises
        ------
        TypeError
            If scheme type is not gem.engine.lib.Configuration
            If logger type is not logging.Logger
        """

        if not isinstance(db_path, Path):
            raise TypeError("Expected %s type for db_path, get %s" % (
                "pathlib.Path", type(db_path)))

        if type(configuration) is not Configuration:
            raise TypeError("Expected %s type for configuration, get %s" % (
                "gem.engine.lib.Configuration", type(configuration)))

        if type(logger) is not Logger:
            raise TypeError("Expected %s type for logger, get %s" % (
                "logging.Logger", type(logger)))

        # ------------------------------------
        #   Variables
        # ------------------------------------

        self.path = db_path.expanduser()

        self.configuration = configuration

        self.logger = logger

        self.sql_types = {
            "NULL": None,
            "BOOL": int,
            "INTEGER": int,
            "REAL": float,
            "TEXT": str,
            "BLOB": memoryview
        }

        # ------------------------------------
        #   Intialization
        # ------------------------------------

        if not self.path.exists():
            for table in self.configuration.sections():
                self.create_table(table)

        else:
            tables = self.select("sqlite_master", ["name"], {"type": "table"})
            if not type(tables) is list:
                tables = [tables]

            tables = list(set(self.configuration.sections()) - set(tables))

            for table in tables:
                self.create_table(table)

    def __generate_request(self, table, data):
        """ Generate a request to database

        This function generate a correct sql request from data dict which use
        keys as columns.

        Parameters
        ----------
        table : str
            Table name
        data : dict
            Columns keys with values

        Returns
        -------
        list
            Request strings list
        """

        values = list()

        if self.configuration.has_section(table):

            for column, value in data.items():
                sql_type = self.configuration.get(table, column).split()[0]

                if sql_type.upper() in list(self.sql_types.keys()):

                    if self.sql_types.get(sql_type.upper()) is str:
                        values.append("%s = \"%s\"" % (
                            str(column), str(data.get(column))))
                    else:
                        values.append("%s = %s" % (
                            str(column), str(data.get(column))))

        else:
            for column, value in data.items():
                values.append("%s = \"%s\"" % (
                    str(column), str(data.get(column))))

        return values

    def create_table(self, table):
        """ Create a new table into database

        Parameters
        ----------
        table : str
            Table name
        """

        request = list()

        database = sqlite3.connect(str(self.path))
        cursor = database.cursor()

        for option in self.configuration.items(table):
            request.append(" ".join(option))

        try:
            with database:
                cursor.execute("CREATE TABLE %(table)s (%(data)s);" % {
                    "table": table,
                    "data": ", ".join(request)
                })

        except Exception as error:
            self.logger.critical(str(error))

        database.commit()
        cursor.close()

    def rename_table(self, table, name):
        """ Rename a table from database

        Parameters
        ----------
        table : str
            Previous table name
        name : str
            New table name
        """

        database = sqlite3.connect(str(self.path))
        cursor = database.cursor()

        try:
            cursor.execute("ALTER TABLE %(table)s RENAME TO %(name)s;" % {
                "table": table,
                "name": name
            })

        except Exception as error:
            self.logger.critical(str(error))

        database.commit()
        cursor.close()

    def remove_table(self, table):
        """ Remove a table from database

        Parameters
        ----------
        table : str
            Table name
        """

        database = sqlite3.connect(str(self.path))
        cursor = database.cursor()

        try:
            cursor.execute("DROP TABLE IF EXISTS %(table)s;" % {
                "table": table
            })

        except Exception as error:
            self.logger.critical(str(error))

        database.commit()
        cursor.close()

    def add_column(self, table, name, sql_type):
        """ Add a new column into database

        Parameters
        ----------
        table : str
            Table name
        name : str
            Column name
        sql_type : str
            Column type
        """

        database = sqlite3.connect(str(self.path))
        cursor = database.cursor()

        try:
            with database:
                cursor.execute(
                    "ALTER TABLE %(table)s ADD COLUMN %(name)s %(type)s;" % {
                        "table": table,
                        "name": name,
                        "type": sql_type
                    })

        except Exception as error:
            self.logger.critical(str(error))

        database.commit()
        cursor.close()

    def get_columns(self, table):
        """ Get all the columns from database

        Parameters
        ----------
        table : str
            Table name

        Returns
        -------
        list
            Columns list
        """

        columns = list()

        database = sqlite3.connect(str(self.path))
        cursor = database.cursor()

        try:
            with database:
                request = cursor.execute("PRAGMA table_info(%(table)s);" % {
                    "table": table
                })

                for data in request.fetchall():
                    columns.append(data[1])

        except Exception as error:
            self.logger.critical(str(error))

        database.commit()
        cursor.close()

        return columns

    def insert(self, table, data):
        """ Insert a new row into database

        This function insert a new row into database from data dict which use
        keys as columns.

        Parameters
        ----------
        table : str
            Table name
        data : dict
            Columns keys and values
        """

        database = sqlite3.connect(str(self.path))
        cursor = database.cursor()

        try:
            values, columns = list(), list()

            for column in list(data.keys()):
                sql_type = self.configuration.get(table, column).split()[0]

                if sql_type.upper() in list(self.sql_types.keys()):

                    if self.sql_types.get(sql_type.upper()) is str:
                        if data.get(column) is not None:
                            columns.append(column)
                            values.append("\"%s\"" % str(data.get(column)))

                    elif data.get(column) is not None:
                        columns.append(column)
                        values.append(str(data.get(column)))

            with database:
                cursor.execute(
                    "INSERT INTO %(table)s (%(columns)s) VALUES (%(data)s);"
                    "" % {
                        "table": table,
                        "columns": ", ".join(columns),
                        "data": ", ".join(values)
                    })

        except Exception as error:
            self.logger.critical(str(error))

        database.commit()
        cursor.close()

    def update(self, table, data, where):
        """ Update a row from database

        This function update a row from database with data and where dict which
        use keys as columns.

        Parameters
        ----------
        table : str
            Table name
        data : dict
            Columns keys and values
        where : dict
            Request conditions
        """

        database = sqlite3.connect(str(self.path))
        cursor = database.cursor()

        try:
            values = self.__generate_request(table, data)
            conditions = self.__generate_request(table, where)

            with database:
                cursor.execute(
                    "UPDATE %(table)s SET %(data)s WHERE %(where)s;" % {
                        "table": table,
                        "data": ", ".join(values),
                        "where": " AND ".join(conditions)
                    })

        except Exception as error:
            self.logger.critical(str(error))

        database.commit()
        cursor.close()

    def select(self, table, columns, where=None):
        """ Get rows from the database

        This function do a request for specific columns from database with
        where dict which use keys as columns.

        Parameters
        ----------
        table : str
            Table name
        columns : list
            Columns name list
        where : dict, optional
            Request conditions (default: None)

        Returns
        -------
        object or None
            Database rows

        Examples
        --------
        >>> Database.get("main", ["age"], {"name": "doe"})
        {'age': 42}
        """

        value = None

        if type(columns) is not list:
            columns = [columns]

        database = sqlite3.connect(str(self.path))
        cursor = database.cursor()

        try:
            if where is None:
                request = cursor.execute(
                    "SELECT %(columns)s FROM %(table)s;" % {
                        "table": table,
                        "columns": ", ".join(columns)
                    })

            else:
                conditions = self.__generate_request(table, where)

                request = cursor.execute(
                    "SELECT %(columns)s FROM %(table)s WHERE %(where)s;" % {
                        "table": table,
                        "columns": ", ".join(columns),
                        "where": " AND ".join(conditions)
                    })

            value = request.fetchall()

            if len(columns) == 1 and '*' not in columns:
                value = [index[0] for index in value]

        except Exception as error:
            self.logger.critical(str(error))

        database.commit()
        cursor.close()

        if value is not None and len(value) == 0:
            return None
        elif value is not None and len(value) == 1:
            return value[0]

        return value

    def remove(self, table, where):
        """ Remove data from database

        This function remove a row from database with where dict which use keys
        as columns.

        Parameters
        ----------
        table : str
            Table name
        where : dict
            Request conditions
        """

        database = sqlite3.connect(str(self.path))
        cursor = database.cursor()

        try:
            with database:
                conditions = self.__generate_request(table, where)

                cursor.execute("DELETE FROM %(table)s WHERE %(where)s;" % {
                    "table": table,
                    "where": " AND ".join(conditions)
                })

        except Exception as error:
            self.logger.critical(str(error))

        database.commit()
        cursor.close()

    def modify(self, table, data, where=None):
        """ Set a specific data in main table

        This function insert or update a row from database with data and where
        dict which use keys as columns.

        Parameters
        ----------
        table : str
            Table name
        data : dict
            Columns keys and values
        where : dict, optional
            Request conditions (Default: None)
        """

        request = self.select(table, list(data.keys()), where)

        if request is None:
            if where is not None:
                data.update(where)

            self.insert(table, data)

        else:
            self.update(table, data, where)

    def get(self, table, where):
        """ Get rows from database

        This function request rows from database with where dict which use keys
        as columns.

        Parameters
        ----------
        table : str
            Table name
        where : dict
            Request conditions

        Returns
        -------
        dict
            Rows data

        Examples
        --------
        >>> Database.get("main", {"name": "doe"})
        {'first_name': 'john', 'age': 42}
        """

        result = None

        values = self.select(table, ['*'], where)

        if values is not None:
            columns = self.get_columns(table)

            result = dict()
            for column in columns:
                value = values[columns.index(column)]
                if value == "None":
                    value = None

                result[column] = value

        return result

    def check_integrity(self):
        """ Check if database respect configuration schema

        Returns
        -------
        bool
            Integrity status
        """

        tables = self.select("sqlite_master", ["name"], {"type": "table"})
        if not type(tables) is list:
            tables = [tables]

        if not sorted(tables) == sorted(self.configuration.sections()):
            return False

        for table in tables:
            columns = self.configuration.options(table)

            if not self.get_columns(table) == columns:
                return False

        return True



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# GEM
from geode_gem.ui.data import Icons
from geode_gem.ui.widgets.window import CommonWindow
from geode_gem.ui.widgets.widgets import ListBoxItem, ScrolledListBox

# GObject
from gi.repository import GdkPixbuf, Gtk, Pango

# Translation
from gettext import gettext as _


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class DNDConsoleDialog(CommonWindow):

    def __init__(self, parent, filepaths):
        """ Constructor

        Parameters
        ----------
        parent : Gtk.Window
            Parent object
        filepaths : dict
            Filepaths storage with corresponding consoles
        """

        classic_theme = False
        if parent is not None:
            classic_theme = parent.use_classic_theme

        CommonWindow.__init__(self,
                              parent,
                              _("Drag & Drop"),
                              Icons.Symbolic.GAMING,
                              classic_theme)

        # ----------------------------------------
        #   Initialize variables
        # ----------------------------------------

        self.api = parent.api

        self.filepaths = filepaths

        self.__notebook_pages = dict()

        # ----------------------------------------
        #   Prepare interface
        # ----------------------------------------

        # Init widgets
        self.__init_widgets()

        # Init signals
        self.__init_signals()

        # Start interface
        self.__start_interface()

    def __init_widgets(self):
        """ Initialize interface widgets
        """

        self.set_size(800, 600)

        self.set_spacing(0)

        self.set_resizable(True)

        # ----------------------------------------
        #   Grid
        # ----------------------------------------

        self.grid.set_spacing(6)

        # ----------------------------------------
        #   Games
        # ----------------------------------------

        self.label_games = Gtk.Label()

        self.notebook_games = Gtk.Notebook()

        self.frame_games = ScrolledListBox()
        self.frame_error = ScrolledListBox()

        # Properties
        self.label_games.set_markup(
            "<b>%s</b>" % _("Drop games"))
        self.label_games.set_hexpand(True)
        self.label_games.set_use_markup(True)
        self.label_games.set_single_line_mode(True)
        self.label_games.set_halign(Gtk.Align.CENTER)
        self.label_games.set_ellipsize(Pango.EllipsizeMode.END)

        # ----------------------------------------
        #   Options
        # ----------------------------------------

        self.label_options = Gtk.Label()

        self.frame_options = Gtk.Frame()
        self.scroll_options = Gtk.ScrolledWindow()
        self.listbox_options = Gtk.ListBox()

        self.widget_copy = ListBoxItem()
        self.switch_copy = Gtk.Switch()

        self.widget_replace = ListBoxItem()
        self.switch_replace = Gtk.Switch()

        self.widget_create = ListBoxItem()
        self.switch_create = Gtk.Switch()

        # Properties
        self.label_options.set_markup(
            "<b>%s</b>" % _("Available actions"))
        self.label_options.set_margin_top(12)
        self.label_options.set_hexpand(True)
        self.label_options.set_use_markup(True)
        self.label_options.set_single_line_mode(True)
        self.label_options.set_halign(Gtk.Align.CENTER)
        self.label_options.set_ellipsize(Pango.EllipsizeMode.END)

        self.scroll_options.set_propagate_natural_height(True)

        self.listbox_options.set_activate_on_single_click(True)
        self.listbox_options.set_selection_mode(Gtk.SelectionMode.NONE)

        self.widget_copy.set_widget(self.switch_copy)
        self.widget_copy.set_option_label(
            _("Copy files"))
        self.widget_copy.set_description_label(
            _("Make copy of dropped files instead of moving them"))

        self.widget_replace.set_widget(self.switch_replace)
        self.widget_replace.set_option_label(
            _("Replace files"))
        self.widget_replace.set_description_label(
            _("Replace existing files on disk"))

        self.widget_create.set_widget(self.switch_create)
        self.widget_create.set_option_label(
            _("Generate subdirectory"))
        self.widget_create.set_description_label(
            _("Create consoles games subdirectory if missing"))

        # ----------------------------------------
        #   Integrate widgets
        # ----------------------------------------

        self.listbox_options.add(self.widget_copy)
        self.listbox_options.add(self.widget_replace)
        self.listbox_options.add(self.widget_create)

        self.scroll_options.add(self.listbox_options)
        self.frame_options.add(self.scroll_options)

        self.grid.pack_start(self.label_games, False, False, 0)
        self.grid.pack_start(self.notebook_games, True, True, 0)
        self.grid.pack_start(self.label_options, False, False, 0)
        self.grid.pack_start(self.frame_options, False, False, 0)

    def __init_signals(self):
        """ Initialize widgets signals
        """

        self.frame_games.listbox.connect(
            "row-activated", self.on_activate_listboxrow)

        self.listbox_options.connect(
            "row-activated", self.on_activate_listboxrow)

    def __start_interface(self):
        """ Load data and start interface
        """

        self.add_button(_("Cancel"), Gtk.ResponseType.CLOSE)
        self.add_button(_("Accept"), Gtk.ResponseType.APPLY, Gtk.Align.END)

        self.set_response_sensitive(Gtk.ResponseType.APPLY, False)

        self.switch_copy.set_active(True)
        self.switch_create.set_active(True)

        # ----------------------------------------
        #   Retrieve data
        # ----------------------------------------

        self.__notebook_pages.clear()

        for path, consoles in sorted(self.filepaths.items()):
            row = DNDGameRow(self.main_parent, path, consoles)

            if len(consoles) == 0:
                self.frame_error.listbox.add(row)

            elif len(consoles) == 1:
                self.frame_games.listbox.add(row)

            else:
                extension = str()

                # Only retrieve extensions and not part of the name
                for subextension in path.suffixes:
                    if subextension not in path.stem:
                        extension += subextension.lower()

                # Generate a tab for the new extension
                if extension not in self.__notebook_pages.keys():
                    self.__notebook_pages[extension] = \
                        DNDSelector(self, consoles)

                self.__notebook_pages[extension].listbox_manual.add(row)

                row.set_console(self.__notebook_pages[extension].get_console())

        # ----------------------------------------
        #   Show notebook tabs
        # ----------------------------------------

        if len(self.frame_games.listbox) > 0:
            self.notebook_games.append_page(
                self.frame_games,
                Gtk.Image.new_from_icon_name(
                    Icons.Symbolic.OK, Gtk.IconSize.MENU))

            self.set_response_sensitive(Gtk.ResponseType.APPLY, True)

        for label, widget in self.__notebook_pages.items():
            self.notebook_games.append_page(widget, Gtk.Label.new(label))

            self.set_response_sensitive(Gtk.ResponseType.APPLY, True)

        if len(self.frame_error.listbox) > 0:
            self.notebook_games.append_page(
                self.frame_error,
                Gtk.Image.new_from_icon_name(
                    Icons.Symbolic.ERROR, Gtk.IconSize.MENU))

    def get_data(self):
        """ Retrieve validate dropped files

        Returns
        -------
        dict
            Validate files as dict structure
        """

        data = dict()

        for row in self.frame_games.listbox:
            data[row.path] = row.console

        for widget in self.__notebook_pages.values():
            for row in widget.listbox_manual:
                data[row.path] = row.console

        return data

    def get_options(self):
        """ Retrieve options values

        Returns
        -------
        dict
            Options values as dict structure
        """

        return {
            "copy": self.switch_copy.get_active(),
            "replace": self.switch_replace.get_active(),
            "create": self.switch_create.get_active()
        }


class DNDSelector(Gtk.Box):

    def __init__(self, parent, consoles):
        """ Constructor

        Parameters
        ----------
        parent : gem.ui.dialog.DNDConsoleDialog
            Parent instance
        consoles : list
            Consoles data list
        """

        Gtk.Box.__init__(self, orientation=Gtk.Orientation.VERTICAL)

        # ----------------------------------------
        #   Variables
        # ----------------------------------------

        self.dialog = parent

        self.consoles = consoles

        # ----------------------------------------
        #   Prepare interface
        # ----------------------------------------

        # Init widgets
        self.__init_widgets()

        # Init packing
        self.__init_packing()

        # Init signals
        self.__init_signals()

        # Start interface
        self.__start_interface()

    def __init_widgets(self):
        """ Initialize interface widgets
        """

        # ----------------------------------------
        #   Automatic selection
        # ----------------------------------------

        self.frame_automatic_description = ScrolledListBox()
        self.item_automatic_description = ListBoxItem.new(
            _("Automatic selection"))

        self.grid_consoles = Gtk.Box()

        self.store_consoles = Gtk.ListStore(GdkPixbuf.Pixbuf, str, object)
        self.combo_consoles = Gtk.ComboBox()
        self.cell_consoles_icon = Gtk.CellRendererPixbuf()
        self.cell_consoles_text = Gtk.CellRendererText()

        self.image_consoles = Gtk.Image()
        self.button_consoles = Gtk.Button()

        # Properties
        self.item_automatic_description.set_widget(self.grid_consoles)
        self.item_automatic_description.set_description_label(
            _("Select this console for every rows"))

        self.grid_consoles.set_spacing(12)

        self.combo_consoles.set_id_column(1)
        self.combo_consoles.set_hexpand(True)
        self.combo_consoles.set_halign(Gtk.Align.FILL)
        self.combo_consoles.set_model(self.store_consoles)
        self.combo_consoles.pack_start(self.cell_consoles_icon, False)
        self.combo_consoles.pack_start(self.cell_consoles_text, True)
        self.combo_consoles.add_attribute(self.cell_consoles_icon, "pixbuf", 0)
        self.combo_consoles.add_attribute(self.cell_consoles_text, "text", 1)

        self.image_consoles.set_from_icon_name(
            Icons.Symbolic.REFRESH, Gtk.IconSize.BUTTON)
        self.button_consoles.add(self.image_consoles)
        self.button_consoles.set_tooltip_text(
            _("Use this console to each games in the list above"))
        self.button_consoles.get_style_context().add_class(
            Gtk.STYLE_CLASS_SUGGESTED_ACTION)

        # ----------------------------------------
        #   Manual selection
        # ----------------------------------------

        self.scroll_manual = Gtk.ScrolledWindow()
        self.listbox_manual = Gtk.ListBox()

        # Properties
        self.scroll_manual.set_propagate_natural_height(True)

        self.listbox_manual.set_activate_on_single_click(True)
        self.listbox_manual.set_selection_mode(Gtk.SelectionMode.NONE)

    def __init_packing(self):
        """ Initialize widgets packing in main window
        """

        self.grid_consoles.pack_start(self.combo_consoles, True, True, 0)
        self.grid_consoles.pack_start(self.button_consoles, False, False, 0)

        self.frame_automatic_description.listbox.add(
            self.item_automatic_description)

        self.scroll_manual.add(self.listbox_manual)

        self.pack_start(self.frame_automatic_description, False, False, 0)
        self.pack_start(Gtk.Separator(), False, False, 0)
        self.pack_start(self.scroll_manual, True, True, 0)

    def __init_signals(self):
        """ Initialize widgets signals
        """

        self.frame_automatic_description.listbox.connect(
            "row-activated", self.dialog.on_activate_listboxrow)

        self.button_consoles.connect(
            "clicked", self.__on_select_console)

        self.listbox_manual.connect(
            "row-activated", self.dialog.on_activate_listboxrow)

    def __start_interface(self):
        """ Load data and start interface
        """

        for console in self.consoles:

            icon = self.dialog.parent.get_pixbuf_from_cache(
                "consoles", 24, console.id, console.icon)

            if icon is None:
                icon = self.dialog.parent.icons_blank.get(24)

            self.store_consoles.append([icon, console.name, console])

        self.combo_consoles.set_active(0)

    def __on_select_console(self, widget, *args):
        """ Change every consoles based on combobox selection

        Parameters
        ----------
        widget : Gtk.Widget
            Object which receive signal
        """

        console = self.get_console()

        for row in self.listbox_manual:
            row.set_console(console)

    def get_console(self):
        """ Retrieve current selected console

        Returns
        -------
        gem.engine.console.Console
            Console instance
        """

        return self.store_consoles.get_value(
            self.combo_consoles.get_active_iter(), 2)


class DNDGameRow(ListBoxItem):

    def __init__(self, parent, path, consoles):
        """ Constructor

        Parameters
        ----------
        parent : Gtk.Window
            Parent object
        path : pathlib.Path
            Game filepath
        consoles : list
            Consoles instances list
        """

        ListBoxItem.__init__(self)

        # ----------------------------------------
        #   Initialize variables
        # ----------------------------------------

        self.parent = parent

        self.path = path
        self.consoles = consoles

        self.console = None

        # ----------------------------------------
        #   Prepare interface
        # ----------------------------------------

        # Init widgets
        self.__init_widgets()

        # Start interface
        self.__start_interface()

    def __init_widgets(self):
        """ Initialize interface widgets
        """

        # ----------------------------------------
        #   Console icon
        # ----------------------------------------

        self.image_console = Gtk.Image()

        # ----------------------------------------
        #   Console selector
        # ----------------------------------------

        self.menu_console = Gtk.Menu()

        self.button_console = Gtk.MenuButton()

        # Properties
        self.button_console.set_popup(self.menu_console)

    def __start_interface(self):
        """ Load data and start interface
        """

        self.set_option_label(self.path.name)

        # No console available
        if len(self.consoles) == 0:
            self.set_description_label(_("No corresponding console"))

        # One console available
        elif len(self.consoles) == 1:
            self.console = self.consoles[0]

            self.set_description_label(str(self.console.path))

            self.set_icon(self.console)

            self.set_widget(self.image_console)

        # Multi-consoles availables
        else:
            self.set_widget(self.button_console)

            self.button_console.set_image(self.image_console)

            for console in self.consoles:
                item = Gtk.MenuItem.new_with_label(console.name)
                item.connect("activate", self.__on_select_console)

                setattr(item, "console", console)

                self.menu_console.append(item)

            self.menu_console.show_all()

    def __on_select_console(self, widget, *args):
        """ Select a specific console for current row

        Parameters
        ----------
        widget : Gtk.Widget
            Object which receive signal
        """

        self.set_console(widget.console)

    def set_icon(self, console):
        """ Update icon image based on console data

        Parameters
        ----------
        console : gem.engine.console.Console
            Console instance
        """

        icon = self.parent.get_pixbuf_from_cache(
            "consoles", 24, console.id, console.icon)

        if icon is None:
            icon = self.parent.icons_blank.get(24)

        self.image_console.set_from_pixbuf(icon)

    def set_console(self, console):
        """ Set selected console

        Paremeters
        ----------
        console : gem.engine.console.Console
            Console instance
        """

        self.console = console

        self.set_icon(console)

        self.set_description_label(str(console.path))



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# Filesystem
from pathlib import Path

# ------------------------------------------------------------------------------
#   Modules - XDG
# ------------------------------------------------------------------------------

try:
    from xdg.BaseDirectory import xdg_data_home
    from xdg.BaseDirectory import xdg_cache_home
    from xdg.BaseDirectory import xdg_config_home

except ImportError:
    from os import environ

    if "XDG_DATA_HOME" in environ:
        xdg_data_home = Path(environ["XDG_DATA_HOME"]).expanduser()
    else:
        xdg_data_home = Path.home().joinpath(".local", "share")

    if "XDG_CACHE_HOME" in environ:
        xdg_cache_home = Path(environ["XDG_CACHE_HOME"]).expanduser()
    else:
        xdg_cache_home = Path.home().joinpath(".cache")

    if "XDG_CONFIG_HOME" in environ:
        xdg_config_home = Path(environ["XDG_CONFIG_HOME"]).expanduser()
    else:
        xdg_config_home = Path.home().joinpath(".config")


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class Metadata:
    """ Store application informations
    """


class Folders:
    """ Store application paths
    """

    APPLICATIONS = Path(xdg_data_home, "applications").expanduser()

    class Default:
        """ Store default application paths
        """

        CACHE = Path(xdg_cache_home).expanduser()
        CONFIG = Path(xdg_config_home).expanduser()
        LOCAL = Path(xdg_data_home).expanduser()


class Icons:
    """ Store icons name
    """

    class Size:
        """ Store icons cache subdirectories
        """

    class Symbolic:
        """ Store symbolic icons name

        Store icon symbolic version which are available in GNOME icons theme
        gnome-icon-theme-symbolic
        """


class Columns:
    """ Store games views informations
    """

    class Key:
        """ Store view flags name
        """

        List = "list"
        Grid = "grid"

    class List:
        """ Store treeview model column index
        """

    class Grid:
        """ Store grid model column index
        """



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# Filesystem
from pathlib import Path

# GEM
from geode_gem.engine.utils import generate_identifier

from geode_gem.ui.data import Icons
from geode_gem.ui.utils import replace_for_markup
from geode_gem.ui.widgets.window import CommonWindow
from geode_gem.ui.widgets.widgets import ListBoxItem

# GObject
from gi.repository import Gtk, Pango

# Translation
from gettext import gettext as _


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class DuplicateDialog(CommonWindow):

    def __init__(self, parent, game):
        """ Constructor

        Parameters
        ----------
        parent : Gtk.Window
            Parent object
        game : gem.engine.game.Game
            Game object instance
        """

        classic_theme = False
        if parent is not None:
            classic_theme = parent.use_classic_theme

        CommonWindow.__init__(self,
                              parent,
                              _("Duplicate a game"),
                              Icons.Symbolic.COPY,
                              classic_theme)

        # ------------------------------------
        #   Initialize variables
        # ------------------------------------

        self.api = parent.api

        self.game = game

        # ------------------------------------
        #   Prepare interface
        # ------------------------------------

        # Init widgets
        self.__init_widgets()

        # Init signals
        self.__init_signals()

        # Start interface
        self.__start_interface()

    def __init_widgets(self):
        """ Initialize interface widgets
        """

        self.set_size(640, 480)

        self.set_spacing(6)

        self.set_resizable(True)

        # ------------------------------------
        #   Grid
        # ------------------------------------

        self.grid_switch = Gtk.Grid()

        # Properties
        self.grid_switch.set_column_spacing(12)
        self.grid_switch.set_row_spacing(6)

        # ------------------------------------
        #   Title
        # ------------------------------------

        self.label_title = Gtk.Label()

        # Properties
        self.label_title.set_markup(
            "<span weight='bold' size='large'>%s</span>" % (
                replace_for_markup(self.game.name)))
        self.label_title.set_use_markup(True)
        self.label_title.set_halign(Gtk.Align.CENTER)
        self.label_title.set_ellipsize(Pango.EllipsizeMode.END)

        # ------------------------------------
        #   Filename
        # ------------------------------------

        self.label_name = Gtk.Label()

        self.entry_name = Gtk.Entry()

        # Properties
        self.label_name.set_markup("<b>%s</b>" % _("New filename"))
        self.label_name.set_margin_top(12)
        self.label_name.set_hexpand(True)
        self.label_name.set_use_markup(True)
        self.label_name.set_single_line_mode(True)
        self.label_name.set_halign(Gtk.Align.CENTER)
        self.label_name.set_ellipsize(Pango.EllipsizeMode.END)

        self.entry_name.set_hexpand(True)
        self.entry_name.set_icon_from_icon_name(
            Gtk.EntryIconPosition.PRIMARY, Icons.Symbolic.SAVE)
        self.entry_name.set_icon_from_icon_name(
            Gtk.EntryIconPosition.SECONDARY, Icons.Symbolic.ERROR)

        # ------------------------------------
        #   Optional data
        # ------------------------------------

        self.label_data = Gtk.Label()

        self.frame_options = Gtk.Frame()
        self.scroll_options = Gtk.ScrolledWindow()
        self.listbox_options = Gtk.ListBox()

        self.widget_database = ListBoxItem()
        self.switch_database = Gtk.Switch()

        self.widget_savestate = ListBoxItem()
        self.switch_savestate = Gtk.Switch()

        self.widget_screenshot = ListBoxItem()
        self.switch_screenshot = Gtk.Switch()

        self.widget_note = ListBoxItem()
        self.switch_note = Gtk.Switch()

        self.widget_memory = ListBoxItem()
        self.switch_memory = Gtk.Switch()

        # Properties
        self.label_data.set_markup(
            "<b>%s</b>" % _("Optional data to duplicate"))
        self.label_data.set_margin_top(12)
        self.label_data.set_hexpand(True)
        self.label_data.set_use_markup(True)
        self.label_data.set_single_line_mode(True)
        self.label_data.set_halign(Gtk.Align.CENTER)
        self.label_data.set_ellipsize(Pango.EllipsizeMode.END)

        self.listbox_options.set_activate_on_single_click(True)
        self.listbox_options.set_selection_mode(
            Gtk.SelectionMode.NONE)

        self.widget_database.set_widget(self.switch_database)
        self.widget_database.set_option_label(
            _("Database"))
        self.widget_database.set_description_label(
            _("Duplicate game data from database"))

        self.widget_savestate.set_widget(self.switch_savestate)
        self.widget_savestate.set_option_label(
            _("Savestates"))
        self.widget_savestate.set_description_label(
            _("Duplicate savestates files"))

        self.widget_screenshot.set_widget(self.switch_screenshot)
        self.widget_screenshot.set_option_label(
            _("Screenshots"))
        self.widget_screenshot.set_description_label(
            _("Duplicate screenshots files"))

        self.widget_note.set_widget(self.switch_note)
        self.widget_note.set_option_label(
            _("Note"))
        self.widget_note.set_description_label(
            _("Duplicate game note file"))

        self.widget_memory.set_widget(self.switch_memory)
        self.widget_memory.set_option_label(
            _("Flash memory"))
        self.widget_memory.set_description_label(
            _("Duplicate flash memory file"))

        # ------------------------------------
        #   Integrate widgets
        # ------------------------------------

        self.listbox_options.add(self.widget_database)
        self.listbox_options.add(self.widget_savestate)
        self.listbox_options.add(self.widget_screenshot)
        self.listbox_options.add(self.widget_note)

        if self.main_parent.check_gba_game_use_mednafen(self.game):
            self.listbox_options.add(self.widget_memory)

        self.scroll_options.add(self.listbox_options)

        self.frame_options.add(self.scroll_options)

        self.pack_start(self.label_title, False, False)
        self.pack_start(self.label_name, False, False)
        self.pack_start(self.entry_name, False, False)
        self.pack_start(self.label_data, False, False)
        self.pack_start(self.frame_options)

    def __init_signals(self):
        """ Initialize widgets signals
        """

        self.entry_name.connect("changed", self.check_filename)

        self.listbox_options.connect(
            "row-activated", self.on_activate_listboxrow)

    def __start_interface(self):
        """ Load data and start interface
        """

        self.add_button(_("Cancel"), Gtk.ResponseType.CANCEL)
        self.add_button(_("Accept"), Gtk.ResponseType.APPLY, Gtk.Align.END)

        self.entry_name.set_text(self.game.path.stem)

        self.check_filename()

    def get_data(self):
        """ Retrieve data to duplicate from user choices

        Returns
        -------
        dict
            Data to duplicate
        """

        # Retrieve specified name
        name = self.entry_name.get_text().strip()

        if len(name) > 0:
            name += self.game.extension

            # Generate file path
            filepath = self.game.path.parent.joinpath(name)
            filename = filepath.stem

            data = {
                "paths": list(),
                "filepath": filepath,
                "database": False
            }

            # ------------------------------------
            #   Game file
            # ------------------------------------

            data["paths"].append((self.game.path, filepath))

            # ------------------------------------
            #   Savestates
            # ------------------------------------

            if self.switch_savestate.get_active():

                for path in self.game.savestates:
                    new_path = Path(
                        str(path).replace(self.game.path.stem, filename))

                    data["paths"].append((path, new_path))

            # ------------------------------------
            #   Screenshots
            # ------------------------------------

            if self.switch_screenshot.get_active():

                for path in self.game.screenshots:
                    new_path = Path(
                        str(path).replace(self.game.path.stem, filename))

                    data["paths"].append((path, new_path))

            # ------------------------------------
            #   Notes
            # ------------------------------------

            if self.switch_note.get_active():
                path = self.api.get_local("notes", f"{self.game.id}.txt")

                if path.exists():
                    data["paths"].append((path, self.main_parent.api.get_local(
                        "notes", generate_identifier(filename) + ".txt")))

            # ------------------------------------
            #   Memory type
            # ------------------------------------

            if self.switch_memory.get_active():
                path = \
                    self.main_parent.get_mednafen_memory_type_file(self.game)

                if path.exists():
                    data["paths"].append(
                        (path, path.replace(self.game.path.stem, filename)))

            # ------------------------------------
            #   Database
            # ------------------------------------

            if self.switch_database.get_active():
                data["database"] = True

            return data

        return None

    def check_filename(self, *args):
        """ Check filename in game folder to detect if a file already exists
        """

        icon = None
        status = True

        # Retrieve specified name
        name = self.entry_name.get_text().strip()

        if len(name) > 0:
            name += self.game.extension

            # Generate file path
            filepath = self.game.path.parent.joinpath(name)

            # Cannot replace an existing file
            if filepath.exists():
                icon = Icons.Symbolic.ERROR
                status = False

        self.entry_name.set_icon_from_icon_name(
            Gtk.EntryIconPosition.SECONDARY, icon)
        self.set_response_sensitive(
            Gtk.ResponseType.APPLY, status)



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# GEM
from geode_gem.ui.data import Icons
from geode_gem.ui.utils import replace_for_markup
from geode_gem.ui.widgets.window import CommonWindow
from geode_gem.ui.widgets.widgets import ListBoxItem

# GObject
from gi.repository import Gtk, Pango

# Translation
from gettext import gettext as _


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class MaintenanceDialog(CommonWindow):

    def __init__(self, parent, game):
        """ Constructor

        Parameters
        ----------
        parent : Gtk.Window
            Parent object
        game : gem.engine.game.Game
            Game object
        """

        classic_theme = False
        if parent is not None:
            classic_theme = parent.use_classic_theme

        CommonWindow.__init__(self,
                              parent,
                              _("Maintenance"),
                              Icons.Symbolic.SYSTEM,
                              classic_theme)

        # ------------------------------------
        #   Variables
        # ------------------------------------

        self.api = parent.api

        self.game = game

        # ------------------------------------
        #   Prepare interface
        # ------------------------------------

        # Init widgets
        self.__init_widgets()

        # Init signals
        self.__init_signals()

        # Start interface
        self.__start_interface()

    def __init_widgets(self):
        """ Initialize interface widgets
        """

        self.set_size(640, 480)

        self.set_spacing(6)

        self.set_resizable(True)

        # ------------------------------------
        #   Grid
        # ------------------------------------

        self.grid_switch = Gtk.Grid()

        # Properties
        self.grid_switch.set_column_spacing(12)
        self.grid_switch.set_row_spacing(6)

        # ------------------------------------
        #   Title
        # ------------------------------------

        self.label_title = Gtk.Label()

        # Properties
        self.label_title.set_markup(
            "<span weight='bold' size='large'>%s</span>" % (
                replace_for_markup(self.game.name)))
        self.label_title.set_use_markup(True)
        self.label_title.set_halign(Gtk.Align.CENTER)
        self.label_title.set_ellipsize(Pango.EllipsizeMode.END)

        # ------------------------------------
        #   Description
        # ------------------------------------

        self.label_description = Gtk.Label()

        # Properties
        self.label_description.set_text(
            _("The following actions are irreversible, be careful!"))
        self.label_description.set_line_wrap(True)
        self.label_description.set_max_width_chars(8)
        self.label_description.set_single_line_mode(False)
        self.label_description.set_justify(Gtk.Justification.FILL)
        self.label_description.set_line_wrap_mode(Pango.WrapMode.WORD)

        # ------------------------------------
        #   Options
        # ------------------------------------

        self.label_data = Gtk.Label()

        self.frame_options = Gtk.Frame()
        self.scroll_options = Gtk.ScrolledWindow()
        self.listbox_options = Gtk.ListBox()

        self.widget_database = ListBoxItem()
        self.switch_database = Gtk.Switch()

        self.widget_log = ListBoxItem()
        self.switch_log = Gtk.Switch()

        self.widget_note = ListBoxItem()
        self.switch_note = Gtk.Switch()

        self.widget_savestate = ListBoxItem()
        self.switch_savestate = Gtk.Switch()

        self.widget_screenshots = ListBoxItem()
        self.switch_screenshots = Gtk.Switch()

        self.widget_cache = ListBoxItem()
        self.switch_cache = Gtk.Switch()

        self.widget_environment = ListBoxItem()
        self.switch_environment = Gtk.Switch()

        # Properties
        self.label_data.set_markup(
            "<b>%s</b>" % _("Available actions"))
        self.label_data.set_margin_top(12)
        self.label_data.set_hexpand(True)
        self.label_data.set_use_markup(True)
        self.label_data.set_single_line_mode(True)
        self.label_data.set_halign(Gtk.Align.CENTER)
        self.label_data.set_ellipsize(Pango.EllipsizeMode.END)

        self.listbox_options.set_activate_on_single_click(True)
        self.listbox_options.set_selection_mode(
            Gtk.SelectionMode.NONE)

        self.widget_database.set_widget(self.switch_database)
        self.widget_database.set_option_label(
            _("Database"))
        self.widget_database.set_description_label(
            _("Clean game data from database"))

        self.widget_log.set_widget(self.switch_log)
        self.widget_log.set_option_label(
            _("Log"))
        self.widget_log.set_description_label(
            _("Delete game log file"))

        self.widget_note.set_widget(self.switch_note)
        self.widget_note.set_option_label(
            _("Note"))
        self.widget_note.set_description_label(
            _("Delete game note file"))

        self.widget_savestate.set_widget(self.switch_savestate)
        self.widget_savestate.set_option_label(
            _("Savestates"))
        self.widget_savestate.set_description_label(
            _("Delete savestates files"))

        self.widget_screenshots.set_widget(self.switch_screenshots)
        self.widget_screenshots.set_option_label(
            _("Screenshots"))
        self.widget_screenshots.set_description_label(
            _("Delete screenshots files"))

        self.widget_cache.set_widget(self.switch_cache)
        self.widget_cache.set_option_label(
            _("Icons cache"))
        self.widget_cache.set_description_label(
            _("Delete generated icons from cache"))

        self.widget_environment.set_widget(self.switch_environment)
        self.widget_environment.set_option_label(
            _("Environment variables"))
        self.widget_environment.set_description_label(
            _("Delete custom game environment variables"))

        # ------------------------------------
        #   Integrate widgets
        # ------------------------------------

        self.listbox_options.add(self.widget_database)
        self.listbox_options.add(self.widget_cache)
        self.listbox_options.add(self.widget_log)
        self.listbox_options.add(self.widget_note)
        self.listbox_options.add(self.widget_savestate)
        self.listbox_options.add(self.widget_screenshots)
        self.listbox_options.add(self.widget_environment)

        self.scroll_options.add(self.listbox_options)

        self.frame_options.add(self.scroll_options)

        self.pack_start(self.label_title, False, False)
        self.pack_start(self.label_description, False, False)
        self.pack_start(self.label_data, False, False)
        self.pack_start(self.frame_options)

    def __init_signals(self):
        """ Initialize widgets signals
        """

        self.listbox_options.connect(
            "row-activated", self.on_activate_listboxrow)

    def __start_interface(self):
        """ Load data and start interface
        """

        self.add_button(_("Cancel"), Gtk.ResponseType.CANCEL)
        self.add_button(_("Apply"), Gtk.ResponseType.APPLY, Gtk.Align.END)

    def get_data(self):
        """ Retrieve data to remove from user choices

        Returns
        -------
        dict
            Data to remove
        """

        data = {
            "paths": list(),
            "log": False,
            "database": False,
            "environment": False
        }

        # ------------------------------------
        #   Log
        # ------------------------------------

        if self.switch_log.get_active():
            path = self.api.get_local("logs", f"{self.game.id}.log")

            if path.exists():
                data["paths"].append(path)

                data["log"] = True

        # ------------------------------------
        #   Note
        # ------------------------------------

        if self.switch_note.get_active():
            path = self.api.get_local("notes", f"{self.game.id}.txt")

            if path.exists():
                data["paths"].append(path)

        # ------------------------------------
        #   Savestates
        # ------------------------------------

        if self.switch_savestate.get_active():
            data["paths"].extend(self.game.savestates)

        # ------------------------------------
        #   Screenshots
        # ------------------------------------

        if self.switch_screenshots.get_active():
            data["paths"].extend(self.game.screenshots)

        # ------------------------------------
        #   Cache
        # ------------------------------------

        if self.switch_cache.get_active():

            for size in ("22x22", "96x96"):
                path = self.main_parent.get_icon_from_cache(
                    "games", size, "%s.png" % self.game.id)

                if path.exists():
                    data["paths"].append(path)

        # ------------------------------------
        #   Environment
        # ------------------------------------

        if self.switch_environment.get_active():
            data["environment"] = True

        # ------------------------------------
        #   Database
        # ------------------------------------

        if self.switch_database.get_active():
            data["database"] = True

        return data
